"""This module contains doctests and unit tests which exercise the crds.uses
module that identifies mappings that reference a file.

XXX IMPORTANT:  crds.uses tests are extremely time consuming, DISABLED
"""
import os, os.path
from pprint import pprint as pp

from crds.core import log, config
from crds import uses
from crds import tests
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

HERE = os.path.dirname(__file__) or "."

def dt_disabled_uses_finaall_mappings_using_reference():
    """
    >>> old_state = test_config.setup()

    >> uses.UsesScript("crds.uses --files v2e20129l_flat.fits")()
    hst.pmap
    hst_0001.pmap
    hst_0002.pmap
    hst_0003.pmap
    hst_0004.pmap
    hst_0005.pmap
    hst_0006.pmap
    hst_cos.imap
    hst_cos_0001.imap
    hst_cos_flatfile.rmap
    hst_cos_flatfile_0002.rmap
    0

    >>> test_config.cleanup(old_state)
    """

def dt_disabled_uses_rmaps():
    """
    >>> old_state = test_config.setup()

    >> uses.UsesScript("crds.uses --files hst_cos_flatfile.rmap hst_acs_darkfile.rmap --include-used")()
    hst_cos_flatfile.rmap hst.pmap
    hst_cos_flatfile.rmap hst_0001.pmap
    hst_cos_flatfile.rmap hst_0002.pmap
    hst_cos_flatfile.rmap hst_cos.imap
    hst_acs_darkfile.rmap hst.pmap
    hst_acs_darkfile.rmap hst_0001.pmap
    hst_acs_darkfile.rmap hst_0002.pmap
    hst_acs_darkfile.rmap hst_0003.pmap
    hst_acs_darkfile.rmap hst_0004.pmap
    hst_acs_darkfile.rmap hst_0005.pmap
    hst_acs_darkfile.rmap hst_0006.pmap
    hst_acs_darkfile.rmap hst_0007.pmap
    hst_acs_darkfile.rmap hst_0008.pmap
    hst_acs_darkfile.rmap hst_0009.pmap
    hst_acs_darkfile.rmap hst_0010.pmap
    hst_acs_darkfile.rmap hst_0011.pmap
    hst_acs_darkfile.rmap hst_0012.pmap
    hst_acs_darkfile.rmap hst_0013.pmap
    hst_acs_darkfile.rmap hst_0014.pmap
    hst_acs_darkfile.rmap hst_acs.imap
    hst_acs_darkfile.rmap hst_acs_0001.imap
    0

    >>> test_config.cleanup(old_state)
    """

def dt_disabled_uses_imap():
    """
    >>> old_state = test_config.setup()

    >> uses.UsesScript("crds.uses --files hst_cos.imap")()
    hst.pmap
    hst_0001.pmap
    hst_0002.pmap
    0

    >>> test_config.cleanup(old_state)
    """

class TestUses(test_config.CRDSTestCase):
    '''
    def test_get_imap_except(self):
        r = rmap.get_cached_mapping("hst.pmap")
        with self.assertRaises(exceptions.CrdsUnknownInstrumentError):
            r.get_imap("foo")
    '''

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""

    import unittest
    suite = unittest.TestLoader().loadTestsFromTestCase(TestUses)
    unittest.TextTestRunner().run(suite)

    from crds.tests import test_uses, tstmod
    return tstmod(test_uses)

if __name__ == "__main__":
    print(tst())



#! /usr/bin/env python

"""pysh is syntactic sugar which simplifies using the Python
subprocess module for writing command line utility scripts that one
might ordinarily write in a UNIX shell, .e.g. bash, ksh, or csh.

While the subprocess module provides a flexible basis for launching
other programs in a wide variety of contexts, it does lack the
simplicity of shell script for basic tasks.  Pysh is simple sugar
intended to glue a limited subset of shell script into Python.

Taking from the shellscript example of the subprocess module
implementation illustrates what pysh is really about:

First,  shellscript might look like this:

   output=`dmesg | grep hda`

This shellscript can be translated into Python subprocess module code
like this:

   p1 = Popen(["dmesg"], stdout=PIPE)
   p2 = Popen(["grep", "hda"], stdin=p1.stdout, stdout=PIPE)
   output = p2.communicate()[0]

But using pysh,  built upon the subprocess module,  one would say this:

   output = out("dmesg | grep hda")

Pysh supports a syntax very similar to UNIX shell substitution which
is interpreted in the scope the pysh subprocess is called from, e.g.

    filename1 = "something.fits"
    out("ls -l ${filename1}")

is functionally equivalent to:

    out("ls -l 'something.fits'")

Note that each substitution value is limited to being a single UNIX word.
"""
import sys
import os
import re
import glob
import os.path
import inspect
import shlex

from io import StringIO

from subprocess import PIPE, STDOUT, Popen

# =========================================================================

__all__ = [
    "sys", "os", "re", "glob",

    "sh", "out", "err", "out_err", "status", "words", "lines", "usage",  "arg",

    "Shell", "pysh_execfile"
]

# =========================================================================

class SubprocessFailure(RuntimeError):
    """A subprocess of this shell failed by returning a non-zero exit status."""

PYSH_DEBUG = None

def set_debug(flag):
    """Set the global debug status for pysh, affecting Shell debug default
    settings.  Nominally set to True in test/debug environments.

    Since setting debug mode changes the default behavior of Shell, it should
    only be used temporarily to ensure
    """
    global PYSH_DEBUG
    old_flag = PYSH_DEBUG
    PYSH_DEBUG = flag
    return old_flag

def get_debug():
    """Get the global debug flag for pysh, changing default behavior for
    test environments.
    """
    return PYSH_DEBUG

# =========================================================================

class Shell:
    """Shell performs sh-like variable substitutions and returns a callable
    object which runs `command` as a subprocess when called.   The shell object
    records the final command line and manages program I/O,  capturing output
    and error status as attributes which can be inspected later.
    """
    def __init__(self, *command, **keys):
        self.status = None  # overridden by __call__
        self._context = keys.pop("context", None)
        self._input = keys.pop("input", None)
        self._use_shell = keys.pop("use_shell", True)
        self._raise_on_error = keys.pop("raise_on_error", get_debug())
        self._trace_commands = keys.pop("trace_commands", False)
        self._trace_exceptions = keys.pop("trace_exceptions", get_debug())
        self._capture_output = keys.pop("capture_output", False)
        self._independent_error = keys.pop("independent_error", False)

        if self._context is None:
            # subclasses, were there any,  would need to pass this in
            # when calling __init__ from the subclass since it would
            # change the depth of the stack and invalidate the below.
            self._context = _context(1)

        command = self._handle_command(*command)

        if self._capture_output:
            if self._independent_error:
                self._popen = Popen(
                    command, shell=self._use_shell, stdin=PIPE, stdout=PIPE, stderr=PIPE, **keys)
            else:
                self._popen = Popen(
                    command, shell=self._use_shell, stdin=PIPE, stdout=PIPE, stderr=STDOUT, **keys)
        else:
            # The real utility of this branch is when you want to see
            # output on your terminal like an ordinary shellscript.
            # The downside is that output is no longer captured.
            self._popen = Popen(command, shell=self._use_shell, stdin=PIPE, **keys)

        self._command = command
        self._keys = keys
        self.out  = "" # string results stdout, stderr
        self.err  = ""

    def _handle_command(self, *command):
        """Expand the shell syntax for strings in `command` relative to this
        Shell's context.
        """
        command = list(command)
        for i, arg in enumerate(command):
            command[i] = expand_vars(arg, self._context)
        return " ".join(command)

    def __call__(self):
        """Execute the commands in this Shell expression and either raise
        an Exception or return the subprocess error status.
        """
        if self._trace_commands:
            print(repr(self), file=sys.stderr)
        self.out, self.err = self._popen.communicate(self._input)
        if sys.version_info >= (3,0,0):
            if self.out:
                self.out = self.out.decode("utf-8")
            if self.err:
                self.err = self.err.decode("utf-8")
        self.status = self._popen.returncode
        if self._raise_on_error and self._popen.returncode:
            if self._trace_exceptions and not self._trace_commands:
                print(repr(self), file=sys.stderr)
            if self._trace_exceptions:
                print("exception stdout:", self.out, file=sys.stderr)
                print("exception stderr:", self.err, file=sys.stderr)
            raise SubprocessFailure("status = " + str(self._popen.returncode))
        else:
            return self._popen.returncode

    def __repr__(self):
        return "Shell(%s, %s)" % (repr(self._command), repr(self._keys),)

# =========================================================================

# handle peeking back up the call tree into the caller's namespace so that
# caller variable names can be substituted using ${} notation.  Also handle
# command line parameters using $* and $1, etc.

def _context(backup):
    """Peek back up the call tree into the namespace of the caller.  Return
    a namespace dictionary combining his globals and locals.  `backup` is the
    number of stack frames to back up relative to the caller of _context().
    """
    frame = sys._getframe(backup+1)
    context = dict()
    context.update(os.environ)
    context.update(frame.f_globals)
    context.update(frame.f_locals)
    args = inspect.getargvalues(frame)[3]
    context.update(args)
    context = {
        key : shlex.quote(val)
        for key, val in context.items()
        if isinstance(val, str)
    }
    return context

# This could also be made to handle simple $N sys.argv
ENV_VAR = re.compile(r"[$]([a-zA-Z_0-9]+)")
ENV_VAR_NUM = re.compile(r"[$]([0-9]+)")
ENV_VAR_CURLY = re.compile(r"[$]{([a-zA-Z_0-9]+)}")
ENV_VAR_STAR = re.compile(r"([$][*])")

def _replace_dollar(match):
    """Return the substitution for a local, global, or environment variable."""
    return "%(" + match.group(1) + ")s"

def _replace_sysarg(match):
    """Return the substitution for the $<n> syntax,  .e.g. $1 for the
    first command line parameter.
    """
    return sys.argv[int(match.group(1))]

def _replace_star(_match):
    """Return the substitution for $*,  i.e. all of the command line
    words except for the program name.
    """
    return " ".join(sys.argv[1:])

def _env_to_percent_var(cmd):
    """Do all the shell syntax substitutions to cmd."""
    cmd = ENV_VAR_NUM.sub(_replace_sysarg, cmd)
    cmd = ENV_VAR_STAR.sub(_replace_star, cmd)
    cmd = ENV_VAR.sub(_replace_dollar, cmd)
    return ENV_VAR_CURLY.sub(_replace_dollar, cmd)

def expand_vars(string, context=None):
    """Expand the shell expressions in `string`,  backing up `context` stack
    frames to find the local/global substitution context.
    """
    if context is None:
        context = _context(2)
    return _env_to_percent_var(string) % context

# =========================================================================

def sh(*command, **keys):
    """Run a subprogram,  inheriting stdout and stderr from the calller, and
    return the program exit status.  Output is not captured.
    If raise_on_error is True,  raise an exception on non-zero program exit.
    """
    shell = Shell(*command, context=_context(1), capture_output=False, **keys)
    shell()
    return shell.status

def _captured_output(*command, **keys):
    """Make a Shell out of `*command` and `**keys` and execute it, capturing
    the output.   Execute the Shell and return it so that various output
    strings or error status can be fetched from attributes as needed.
    """
    shell = Shell(*command, context=_context(2), capture_output=True, **keys)
    shell()
    return shell

def status(*command, **keys):
    """Run a subprogram capturing it's output and return the exit status."""
    return _captured_output(*command, **keys).status

def out(*command, **keys):
    """Run a subprogram and return it's stdout."""
    return _captured_output(*command, **keys).out

def err(*command, **keys):
    """Run a subprogram and return it's stderr."""
    return _captured_output(*command, **keys).err

def out_err(*command, **keys):
    """Run a subprogram and return it's combined/interleaved stdout and stderr."""
    keys["independent_error"] = False
    return _captured_output(*command, **keys).out

def status_out_err(*command, **keys):
    """Run a subprogram and return it's status and combined/interleaved stdout and stderr."""
    keys["independent_error"] = False
    capture = _captured_output(*command, **keys)
    return capture.status, capture.out

def words(*command, **keys):
    """Return the standard output of `command` split into a sequence
    of words.
    """
    return _captured_output(*command, **keys).out.split()

def lines(*command, **keys):
    """Return the standard output of `command` split into a sequence
    of lines.
    """
    # keys["independent_error"] = False
    return _captured_output(*command, **keys).out.splitlines()

# =========================================================================

# Convenience routines not related to subprocesses here.

def cd(directory):
    """change directory to `directory`"""
    os.chdir(directory)

def fail(*args, **keys):
    """Quit the program,  issuing a message which is the join of *args.
    Use keyword `status` as the program exit code or -1 if unspecified.
    """
    print(" ".join(args), file=keys.pop("file", sys.stderr))
    sys.exit(keys.pop("status", -1))

def usage(description, min_args, max_args=sys.maxsize, help=""):
    """Emit a standard program usage message based on the min and max
    command line parameter counts.   If the program takes at least
    one parameter,  min_args should be 1.   If the program takes at most
    one parameter, max_args should also be 1.
    """
    progname = os.path.split(sys.argv[0])[1]
    if not (min_args <= len(sys.argv)-1 <= max_args) or "--help" in sys.argv:
        fail("\nusage: " + progname + " "  + description + help)

def arg(index, default=None, typecast=str):
    """Ultra-simple convenience function for extracting command line parameters by
    index, optionally using a `default` value if fewer than `index` parameters
    are provided.  The function `typecast` can be set to a python type object
    to cast parameter strings to other python types like int or float.

    $ program  arg1 arg2 arg3

    progname = arg(0)
    arg1 = arg(1, "foo")
    arg2 = arg(2, 57, int)
    arg3 = arg(3, 45.0, float)

    """
    val = sys.argv[index] if len(sys.argv) >= index+1 else default
    return typecast(val)

# =========================================================================

# Code for rewriting a file of pysh-script as an ordinary python module.

def __rewrite_shell_statement(match):
    """Return the substitution for re-writing a line of pysh-script as
    Python.
    """
    return match.group(1) + "sh('''" + match.group(2) + "''')"

SHELL_STATEMENT = re.compile(r"^(\s*)% (.*)$")

def _rewrite_shell_statement(line):
    """Re-write `line` as Python code and return it."""
    return SHELL_STATEMENT.sub(__rewrite_shell_statement, line)

def pysh_execfile(fname, globals=None, locals=None):
    """Re-write pysh-script `fname` as Python and execute it."""
    import tempfile
    with open(fname) as sourcefile:
        lines =  sourcefile.readlines()
    for i, line in enumerate(lines):
        lines[i] = _rewrite_shell_statement(line)
    try:
        (handle, fname) = tempfile.mkstemp()
        for line in lines:
            os.write(handle, line.encode("utf-8"))
    finally:
        os.close(handle)
    try:
        exec(open(fname).read(), globals, locals)
    finally:
        os.remove(fname)

# =========================================================================
# The execution sequence for when pysh is used as a program shell.

if __name__ == "__main__":
    usage("<pysh-scriptname> [parameters...]", 1)
    sys.argv = sys.argv[1:]
    pysh_execfile(sys.argv[0], globals(), locals())



"""This module defines a base class for CRDS command line scripts.

MAYBE integrate rc, environment, and command line parameters.
"""
# -----------------------------------------------------------------------------

import sys
import os
import argparse
import pdb
import cProfile, pstats
import re
from collections import Counter, defaultdict

# -----------------------------------------------------------------------------

from argparse import RawTextHelpFormatter

# -----------------------------------------------------------------------------

from . import log, heavy_client, constants
from . import config, utils, exceptions, rmap
from crds.client import api

# from crds import data_file  :   deferred,  see below

# =============================================================================

def _show_version():
    """Dump CRDS version information and exit python."""
    print(heavy_client.version_info())
    sys.exit(-1)

# =============================================================================

# command line parameter type coercion / verification functions

def dataset(filename):
    """Ensure `filename` names a dataset."""
    from crds import data_file
    if data_file.is_dataset(filename):
        return filename
    else:
        raise ValueError("Parameter", repr(filename),
                         "does not appear to be a dataset filename.")

def reference_file(filename):
    """Ensure `filename` is a reference file."""
    assert config.is_reference(filename), \
        "A reference file is required but got bad type: '%s'" % filename
    return filename

def mapping(filename):
    """Ensure `filename` is a CRDS mapping file."""
    assert config.is_mapping(filename), "A .rmap, .imap, or .pmap file is required but got: '%s'" % filename
    return filename

def mapping_spec(spec):
    """Ensure `spec` is a CRDS mapping specification, a filename or a date based spec."""
    assert config.is_mapping_spec(spec), "A .rmap, .imap, or .pmap file or date base specification is required but got: '%s'" % spec
    return spec

def context_spec(spec):
    """Ensure filename is a .pmap or abstract .pmap like "jwst-edit" or date based context spec."""
    assert config.is_context_spec(spec), \
        "Parameter should be a .pmap or abstract context specifier, not: " + repr(spec)
    return spec

def reference_mapping(filename):
    """Ensure `filename` is a .rmap file."""
    assert filename.endswith(".rmap"), "A .rmap file is required but got: '%s'" % filename
    return filename

#def mapping(filename):
#    """Ensure that `filename` is any known CRDS mapping."""
#    if api.is_known_mapping(filename):
#        return filename
#    else:
#        raise ValueError("Parameter", repr(filename),
#                         "is not a known CRDS mapping.")

def observatory(obs):
    """Verify that `obs` is the name of an observatory and return it."""
    obs = obs.lower()
    assert obs in constants.ALL_OBSERVATORIES, \
        "Unknown observatory " + repr(obs)
    return obs

def nrange(string):
    """Verify a context range expression MIN:MAX and return (MIN, MAX)."""
    assert re.match(r"\d+:\d+", string), \
        "Invalid context range specification " + repr(string)
    rmin, rmax = [int(x) for x in string.split(":")]
    assert 0 <= rmin <= rmax, "Invalid range values"
    return rmin, rmax

def process_key(string):
    """Check the format of a remote process identification  key."""
    assert config.PROCESS_KEY_RE.match(string),  "Invalid format for process key: " + repr(string)
    return string

def user_name(string):
    """Check the format of a server user name string."""
    assert config.USER_NAME_RE.match(string), "Invalid user name " + repr(string)
    return string

# =============================================================================
# =============================================================================

class Script:
    """Base class for CRDS command line scripts with standard properties.

    `args` is either a string of command line parameters or a parameter list of command line words.  If
           defaulted to None then `args` is treated as sys.argv (default argparse handling). Note that `args`
           shoulld include the program name as args[0].  Explicitly specifying `args` is used to
           instantiate a script in code for testing, etc.
    """

    decription = epilog = usage = None
    formatter_class = RawTextHelpFormatter

    def __init__(self, argv=None, parser_pars=None, reset_log=True, print_status=False):
        self.stats = utils.TimingStats()
        self._already_reported_stats = False
        if isinstance(argv, str):
            argv = argv.split()
        elif argv is None:
            argv = sys.argv
        self._argv = argv
        if parser_pars is None:
            parser_pars = {}
        for key in ["description", "epilog", "usage", "formatter_class"]:
            self._add_key(key, parser_pars)
        self.parser = argparse.ArgumentParser(prog=argv[0], **parser_pars)
        self.add_args()
        self.add_standard_args()
        self.args = self.parser.parse_args(argv[1:])
        if self.args.readonly_cache:
            config.set_cache_readonly(True)
        log.set_verbose(log.get_verbose() or self.args.verbosity or self.args.verbose)
        log.set_log_time(config.get_log_time() or self.args.log_time)
        output_cmd = log.info if self.args.dump_cmdline else log.verbose
        output_cmd("Command:", [os.path.basename(argv[0])] + argv[1:], verbosity=30)
        self.print_status = print_status
        self.reset_log = reset_log
        if self.reset_log:
            log.reset()  # reset the infos, warnings, and errors counters as if new commmand line run.
        self._exit_status = None
        self.show_context_resolution = True

    def main(self):
        """Write a main method to perform the actions of the script using self.args."""
        raise NotImplementedError("Script subclasses have to define main().")

    def _main(self):
        """_main() completes any complex generic setup,  like determining contexts, and then
        calls self.main() which does the real work of the script.   _main() defines the full
        call tree of code which is run inside the profiler or debugger.
        """
        self.contexts = self.determine_contexts()
        self._exit_status = self.main()
        self.report_stats()  # here if not called already
        if self.print_status:
            log.standard_status()
        return self._exit_status

    # @data_file.hijack_warnings
    def __call__(self):
        """Run the script's _main() according to command line parameters."""
        try:
            if self.args.debug_traps:
                log.set_exception_trap(False)
            if self.args.version:
                _show_version()
            elif self.args.profile:
                self._profile()
            elif self.args.pdb:
                pdb.runctx("self._main()", globals(), locals())
            else:
                self._main()
        except KeyboardInterrupt as exc:
            if self.args.pdb:
                raise
            else:
                raise KeyboardInterrupt("Interrupted... quitting.") from exc
        return self._exit_status

    @property
    def locator(self):
        """Return the module for observatory specific file locations and plugins functions."""
        return utils.get_locator_module(self.observatory)

    @property
    def obs_pkg(self):
        """Return the package __init__ for observatory specific constants."""
        return utils.get_observatory_package(self.observatory)

    def determine_contexts(self):
        """Return the list of contexts used by this invocation of the script.  Empty for Script."""
        return []

    def add_args(self):
        """Add script-specific argparse add_argument calls here on self.parser"""
        # raise NotImplementedError("Script subclasses have to define add_args().")

    @property
    def readonly_cache(self):
        """Return True of the cache is readonly."""
        return config.get_cache_readonly()

    @property
    @utils.cached
    def observatory(self):
        """Return either the command-line override observatory,  or the one determined
        by the client/server exchange.
        """
        if self.args.jwst:
            return self.set_server("jwst")
        if self.args.hst:
            return self.set_server("hst")
        if self.args.roman:
            return self.set_server("roman")

        obs = config.OBSERVATORY.get()
        if obs != "none":
            return self.set_server(obs.lower())

        url = os.environ.get("CRDS_SERVER_URL", None)
        if url is not None:
            for obs in constants.ALL_OBSERVATORIES:
                if obs in url.lower():
                    return self.set_server(obs)

        files = []
        if hasattr(self, "contexts"):
            files += self.contexts
        if hasattr(self.args, "files"):
            files += self.args.files if self.args.files else []

        for file_ in files:
            if file_.startswith("hst"):
                return self.set_server("hst")
            if file_.startswith("jwst"):
                return self.set_server("jwst")

        for file_ in files:
            with log.verbose_on_exception("Failed file_to_observatory for", repr(file_)):
                return self.set_server(utils.file_to_observatory(file_))

        return api.get_default_observatory()

    @property
    def exit_status(self):
        """Retrieve the script's exit status as it would have been sent to
        sys.exit(...), had the script been run from the commandline."""

        if self._exit_status is None:
            raise RuntimeError("exit_status is not available until script has completed")
        return self._exit_status

    def set_server(self, observatory):
        """Based on `observatory`,  set the CRDS server to an appropriate default,  particularly
        for the case where CRDS_SERVER_URL is not set.
        """
        url = config.get_server_url(observatory)
        if url is not None:
            api.set_crds_server(url)
        return observatory

    def _add_key(self, key, parser_pars):
        """Add any defined class attribute for `key` to dict `parser_pars`."""
        inlined = getattr(self, key, parser_pars)
        if inlined is not None:
            parser_pars[key] = inlined
        return parser_pars

    def add_argument(self, *args, **keys):
        """Add a parser argument."""
        self.parser.add_argument(*args, **keys)

    def get_exclusive_arg_group(self, *args, **keys):
        """Return a mutually exlusive argument group."""
        return self.parser.add_mutually_exclusive_group(*args, **keys)

    def add_standard_args(self):
        """Add standard CRDS command line parameters."""
        self.add_argument("-v", "--verbose",
            help="Set log verbosity to True,  nominal debug level.", action="store_true")
        self.add_argument("--verbosity",
            help="Set log verbosity to a specific level: 0..100.", type=int, default=0)
        self.add_argument("--dump-cmdline", action="store_true",
            help="Dump the command line parameters used to start the script to the log.")
        self.add_argument("-R", "--readonly-cache", action="store_true",
            help="Don't modify the CRDS cache.  Not compatible with options which implicitly modify the cache.")
        self.add_argument('-I', '--ignore-cache', action='store_true', dest="ignore_cache",
                          help="Download required files even if they're already in the cache.")
        self.add_argument("-V", "--version",
            help="Print the software version and exit.", action="store_true")
        self.add_argument("-J", "--jwst", dest="jwst", action="store_true",
            help="Force observatory to JWST for determining header conventions.""")
        self.add_argument("-H", "--hst",  dest="hst", action="store_true",
            help="Force observatory to HST for determining header conventions.""")
        self.add_argument("--roman",  dest="roman", action="store_true",
            help="Force observatory to Roman for determining header conventions.""")
        self.add_argument("--stats", action="store_true",
            help="Track and print timing statistics.")
        self.add_argument("--profile",
            help="Output profile stats to the specified file.", type=str, default="")
        self.add_argument("--log-time", action="store_true",
            help="Add date/time to log messages.")
        self.add_argument("--pdb",
            help="Run under pdb.", action="store_true")
        self.add_argument("--debug-traps",
            help="Bypass exception error message traps and re-raise exception.", action="store_true")

    def print_help(self):
        """Print out command line help."""
        self.parser.print_help()

    def require_server_connection(self):
        """Check a *required* server connection and ERROR/exit if offline."""
        try:
            if not self.server_info.connected:
                raise RuntimeError("Required server connection unavailable.")
        except Exception as exc:
            self.fatal_error("Failed connecting to CRDS server at CRDS_SERVER_URL =",
                             repr(api.get_crds_server()), "::", str(exc))

    @property
    @utils.cached
    def server_info(self):   # see also crds.sync server_info which does not update.
        """Return the server_info dict from the CRDS server *or* cache config for non-networked use where possible."""
        info = heavy_client.get_config_info(self.observatory)
        heavy_client.update_config_info(self.observatory)
        return info

    @property
    @utils.cached
    def bad_files(self):
        """Return the current set of ALL known bad mappings and references, not context-specific."""
        return self.server_info.bad_files_set

    @property
    def default_context(self):
        """Return the default operational .pmap defined by the CRDS server or cache."""
        return self.server_info["operational_context"]

    def get_words(self, word_list):
        """Process a file list,  expanding @-files into corresponding lists of
        files.   Return a flat, depth-first,  file list.
        """
        words = []
        for word in word_list:
            if word.startswith("@"):
                words.extend(self._load_word_list(word[1:]))
            else:
                words.append(word)
        return words # [word.lower() for word in words]

    def _load_word_list(self, at_file):
        """Recursively load an @-file, returning a list of words.
        Any stripped line beginning with # is a comment line to be ignored.
        Any word beginning with @ is a file to load recursively.
        Each line is split into words/words using whitespace.
        """
        words = []
        with open(at_file) as atf:
            for line in atf.readlines():
                word = line.strip()
                if word.startswith("#"):
                    continue
                if word.startswith("@"):
                    more = self._load_word_list(word[1:])
                else:
                    more = word.split()
                words.extend(more)
        return self.get_words(words)   # another pass to fix paths

    def get_files(self, file_list):

        """Expand a list of files by treating any filename beginning with an
        @-sign as a file containing one word per line.
        """
        return self.get_words(file_list)

    @property
    def files(self):
        """Handle @-files and add cache_paths to command line file parameters.
        Nominally self.files are assumed to be references or mappings.  Override locate_file()
        to handle other files.
        """
        if not hasattr(self.args, "files"):
            raise NotImplementedError("Class must implement list of `self.args.files` raw file paths.")
        files1 = self.get_files(self.args.files)
        files2 = []
        for file in files1:
            files2.extend(expand_all_instruments(self.observatory, file))
        return [self.locate_file(fname) for fname in files2]


    #
    # NOTES:
    # crds:// will always mean "inside the cache"
    # ./ will always mean "current directory"
    # pathless files are more ambiguous,  historically in CRDS they mean "in the cache"
    #    but in most/all OSes pathless means "current directory" so concievably could change
    #
    def locate_file(self, filename):
        """Locate file defines how members of the self.args.files list are located.
        The default behavior is to locate CRDS cached files,  either references or mappings.
        This is inappropriate for datasets so in some cases locate_file needs to be overridden.
        Symbolic context names, e.g. hst-operatonal, resolved to literal contexts, e.g. hst_0320.pmap
        """
        filename = config.pop_crds_uri(filename)   # nominally crds://
        filename = self.resolve_context(filename) if config.is_date_based_mapping_spec(filename) else filename
        return config.locate_file(filename, observatory=self.observatory)

    def locate_file_outside_cache(self, filename):
        """This is essentially normal filename syntax,  except crds:// is interpreted to mean
        locate filename inside the CRDS cache.  symbolic context names are also resolved to
        literal context filenames.
        """
        filename2 = config.pop_crds_uri(filename)   # nominally crds://
        filename2 = self.resolve_context(filename2) # e.g. hst-operational -->  hst_0320.pmap
        if filename != filename2:  # Had crds:// or was date based
            return config.locate_file(filename2, self.observatory)
        else:
            if not os.path.dirname(filename2):
                return "./" + filename2
            else:
                return filename2
            # return os.path.abspath(filename2)

    def _profile(self):
        """Run _main() under the Python profiler."""
        if self.args.profile == "console":
            self._console_profile(self._main)
        else:
            cProfile.runctx("self._main()", locals(), locals(), self.args.profile)

    def _console_profile(self, function):
        """Run `function` under the profiler and print results to console."""
        prof = cProfile.Profile()
        prof.enable()
        function()
        prof.disable()
        prof_stats = pstats.Stats(prof).sort_stats("cumulative")
        prof_stats.print_stats(100)

    def report_stats(self):
        """Print out collected statistics."""
        if self.args.stats and not self._already_reported_stats:
            self.stats.report()
            self._already_reported_stats = True

    def increment_stat(self, name, amount=1):
        """Add `amount` to the statistics counter for `name`."""
        self.stats.increment(name, amount)

    def get_stat(self, name):
        """Return statistic `name`."""
        return self.stats.get_stat(name)

    def run(self, *args, **keys):
        """script.run() is the same thing as script() but more explicit."""
        return self.__call__(*args, **keys)

    def resolve_context(self, context):
        """Resolve context spec `context` into a .pmap, .imap, or .rmap filename,  interpreting
        date based specifications against the CRDS server operational context history.
        """
        if isinstance(context, str) and context.lower() == "none":
            return None
        if not config.is_date_based_mapping_spec(context):
            return context
        final_context = heavy_client.get_context_name(self.observatory, context)
        if self.show_context_resolution:
            log.info("Symbolic context", repr(context), "resolves to", repr(final_context))
        return final_context

    def get_conjugates(self, file_list):
        """Given a list of references,  return any GEIS data files associated with them."""
        from crds import data_file
        return [ data_file.get_conjugate(ref) for ref in file_list if data_file.get_conjugate(ref) is not None]

    def get_file_properties(self, filename):
        """Return (instrument, filekind) corresponding to `file`, and '' for none."""
        return utils.get_file_properties(self.observatory, filename)

    def categorize_files(self, filepaths):
        """Organize files in list `filepaths` into a dictionary of lists as follows:
            { (instrument, filekind) : filepath, ... }
        """
        categorized = defaultdict(list)
        for path in filepaths:
            instrument, filekind = self.get_file_properties(path)
            categorized[(instrument, filekind)].append(path)
        return dict(categorized)

    def fatal_error(self, *args, **keys):
        """Issue an error message and terminate the program."""
        log.fatal_error(*args, **keys)

    def dump_files(self, context=None, files=None, ignore_cache=None):
        """Download mapping or reference `files1` with respect to `context`,  tracking stats."""
        if context is None:
            context = self.default_context
        if ignore_cache is None:
            ignore_cache = self.args.ignore_cache
        _localpaths, downloads, nbytes = api.dump_files(
            context, files, ignore_cache=ignore_cache, raise_exceptions=self.args.pdb)
        self.increment_stat("total-files", downloads)
        self.increment_stat("total-bytes", nbytes)

    def dump_mappings(self, mappings, ignore_cache=None):
        """Download all `mappings` and their dependencies if not already cached.."""
        if ignore_cache is None:
            ignore_cache = self.args.ignore_cache
        if not self.server_info.connected:
            log.verbose("Not connected to server. Skipping dump_mappings", mappings, verbosity=55)
            return
        for mapping in mappings:
            _localpaths, downloads, nbytes = api.dump_mappings3(
                mapping, ignore_cache=ignore_cache, raise_exceptions=self.args.pdb)
            self.increment_stat("total-files", downloads)
            self.increment_stat("total-bytes", nbytes)

    def sync_files(self, files, context=None, ignore_cache=None):
        """Like dump_files(),  but dumps recursive closure of any mappings rather than just the listed mapping."""
        mappings = [ os.path.basename(filename)
                     for filename in files if config.is_mapping(filename) ]
        references = [os.path.basename(filename)
                      for filename in files if not config.is_mapping(filename) ]
        if mappings:
            self.dump_mappings(mappings, ignore_cache)
        if references:
            self.dump_files(context, references, ignore_cache)

    def are_all_references(self, files):
        """Return True IFF every file in files is a reference."""
        for filename in files:
            if not config.is_reference(filename):
                return False
        else:
            return True

    def are_all_mappings(self, files):
        """Return True IFF every file in files is a mapping."""
        for filename in files:
            if not config.is_mapping(filename):
                return False
        else:
            return True

# =============================================================================

class UniqueErrorsMixin:
    """This mixin supports tracking certain errors messages."""
    def __init__(self, *args, **keys):

        self.ue_mixin = self.get_empty_mixin()

        # Exception trap context manager for use in "with" blocks
        # trapping exceptions.
        self.error_on_exception = log.exception_trap_logger(
            self.log_and_track_error)  # can be overridden

    def get_empty_mixin(self):
        """Return a bundle of freshly initialized counters and tracking information.   The
        bundle is used to isolate mixin parameters from subclass parameters to prevent
        accidental overrides.
        """
        class Struct:
            pass
        mixin = Struct()
        mixin.messages = {}
        mixin.count = Counter()
        mixin.tracked_errors = 0
        mixin.unique_data_names = set()
        mixin.all_data_names = set()
        mixin.data_names_by_key = defaultdict(list)
        mixin.announce_suppressed = Counter()
        return mixin

    def clear_error_counts(self):
        """Clear the error tracking status by re-initializing/zeroing mixin data structures."""
        self.ue_mixin = self.get_empty_mixin()

    def add_args(self):
        """Add command line parameters to Script arg parser."""
        self.add_argument("--dump-unique-errors", action="store_true",
            help="Record and dump the first instance of each kind of error.")
        self.add_argument("--unique-errors-file",
            help="Write out data names (ids or filenames) for first instance of unique errors to specified file.")
        self.add_argument("--all-errors-file",
            help="Write out all err'ing data names (ids or filenames) to specified file.")
        self.add_argument("--unique-threshold", type=int, default=1,
            help="Only print unique error classes with this many or more instances.")
        self.add_argument("--max-errors-per-class", type=int, default=500, metavar="N",
            help="Only print the first N detailed errors of any particular class.")
        self.add_argument("--unique-delimiter", type=str, default=None,
            help="Use the given delimiter (e.g. semicolon) in tracked error messages to make them amenable to spreadsheets.")

    def log_and_track_error(self, data, instrument, filekind, *params, **keys):
        """Issue an error message and record the first instance of each unique kind of error,  where "unique"
        is defined as (instrument, filekind, msg_text) and omits data id.
        """
        # Always count messages
        self.ue_mixin.tracked_errors += 1
        msg = self.format_prefix(data, instrument, filekind, *params, **keys)
        key = log.format(instrument, filekind.upper(), *params, **keys)
        if key not in self.ue_mixin.messages:
            self.ue_mixin.messages[key] = msg
            self.ue_mixin.unique_data_names.add(data)
        self.ue_mixin.count[key] += 1
        self.ue_mixin.all_data_names.add(data)
        self.ue_mixin.data_names_by_key[key].append(data)
        # Past a certain max,  supress the error log messages.
        if self.ue_mixin.count[key] < self.args.max_errors_per_class:
            log.error(msg)
        else:
            log.increment_errors()
            # Before suppressing,  announce the suppression
            if not self.ue_mixin.announce_suppressed[key]:
                self.ue_mixin.announce_suppressed[key] += 1 # flag
                log.info("Max error count %d exceeded for:" % self.args.max_errors_per_class,
                         key.strip(), "suppressing remaining error messages.")
        return None # for log.exception_trap_logger  --> don't reraise

    def format_prefix(self, data, instrument, filekind, *params, **keys):
        """Create a standard (instrument,filekind,data) prefix for log messages."""
        delim = self.args.unique_delimiter  # for spreadsheets
        data, instrument, filekind = str(data), str(instrument), str(filekind) # squash 2.7 unicode
        if delim:
            return log.format(delim, instrument.upper(), delim, filekind.upper(), delim, data, delim,
                              *params, end="", **keys)
        else:
            return log.format("instrument="+repr(instrument.upper()), "type="+repr(filekind.upper()), "data="+repr(data), ":: ",
                              *params, end="", **keys)

    def dump_unique_errors(self):
        """Print out the first instance of errors recorded by log_and_track_error().  Write out error list files."""
        if self.args.dump_unique_errors:
            if self.args.unique_threshold > 1:
                log.info("Limiting error class reporting to cases with at least",
                         self.args.unique_threshold, "instances.")
            log.info("="*20, "unique error classes", "="*20)
            classes = len(self.ue_mixin.messages)
            for key in sorted(self.ue_mixin.messages):
                if self.ue_mixin.count[key] >= self.args.unique_threshold:
                    log.info("%06d" % self.ue_mixin.count[key], "errors like::", self.ue_mixin.messages[key])
                else:
                    self.drop_error_class(key)
                    classes -= 1
            log.info("All unique error types:", classes)
            log.info("Untracked errors:", log.errors() - self.ue_mixin.tracked_errors)
            log.info("="*20, "="*len("unique error classes"), "="*20)
        if self.args.all_errors_file:
            self.dump_error_data(self.args.all_errors_file, self.ue_mixin.all_data_names)
        if self.args.unique_errors_file:
            self.dump_error_data(self.args.unique_errors_file, self.ue_mixin.unique_data_names)

    def drop_error_class(self, key):
        """Remove the errors classified by `key` from the error classes and counts."""
        for data in self.ue_mixin.data_names_by_key[key]:
            self.ue_mixin.all_data_names = self.ue_mixin.all_data_names - set([data])
            self.ue_mixin.unique_data_names = self.ue_mixin.unique_data_names - set([data])
            self.ue_mixin.count[key] -= 1

    def dump_error_data(self, filename, error_list):
        "Write out list of err'ing filenames or dataset ids to `filename`."""
        with open(filename, "w+") as err_file:
            err_file.write("\n".join(sorted(error_list))+"\n")

# =============================================================================

class ContextsScript(Script):
    """Baseclass for a script proving support for command line specified contexts."""

    def __init__(self, *args, **keys):
        super(ContextsScript, self).__init__(*args, **keys)
        self.contexts = []

    def add_args(self):
        group = self.get_exclusive_arg_group(required=False)
        group.add_argument('--contexts', metavar='CONTEXT', type=mapping_spec, nargs='*',
            help="Specify a list of CRDS mappings to operate on: .pmap, .imap, or .rmap or date-based specification")
        group.add_argument("--range", metavar="MIN:MAX",  type=nrange, dest="range", default=None,
            help='Operate for pipeline context ids (.pmaps) between <MIN> and <MAX>.')
        group.add_argument('--all', action='store_true',
            help='Operate with respect to all known CRDS contexts.')
        group.add_argument('--last-n-contexts', metavar="N", type=int, default=None,
            help='Operate with respect to the last N contexts.')
        group.add_argument("--up-to-context", metavar='CONTEXT', type=mapping_spec, nargs=1, default=None,
            help='Operate on all contexts up to and including the specified context.')
        group.add_argument("--after-context", metavar='CONTEXT', type=mapping_spec, nargs=1, default=None,
            help='Operate on all contexts after and including the specified context.')

        self.add_argument('--include-orphans', action='store_true',
            help='Include reference files not mentioned by any contexts.')

    def determine_contexts(self):
        """Support explicit specification of contexts, context id range, or all."""
        log.verbose("Determining contexts.", verbosity=55)
        if self.args.contexts:
            # permit instrument and reference mappings,  not just pipelines:
            _contexts2 = []
            for ctx in self.args.contexts:
                _contexts2.extend(expand_all_instruments(self.observatory, ctx))
            contexts = []
            for ctx in _contexts2:
                resolved = self.resolve_context(ctx)
                if resolved != 'N/A':
                    contexts.append(resolved)
        elif self.args.all:
            contexts = self._list_mappings("*.pmap")
        elif self.args.last_n_contexts:
            contexts = self._list_mappings("*.pmap")[-self.args.last_n_contexts:]
        elif self.args.range:
            rmin, rmax = self.args.range
            contexts = []
            all_contexts = self._list_mappings("*.pmap")
            for context in all_contexts:
                match = re.match(r"\w+_(\d+).pmap", context)
                if match:
                    serial = int(match.group(1))
                    if rmin <= serial <= rmax:
                        contexts.append(context)
        elif self.args.up_to_context:
            pmaps = self._list_mappings("*.pmap")
            with log.augment_exception("Invalid --up-to-context", repr(self.args.up_to_context[0]), exc_class=exceptions.CrdsError):
                up_to_context = self.resolve_context(self.args.up_to_context[0])
                up_to_ix = pmaps.index(up_to_context)+1
                contexts = pmaps[:up_to_ix]
        elif self.args.after_context:
            pmaps = self._list_mappings("*.pmap")
            with log.augment_exception("Invalid --after-context", repr(self.args.after_context[0]), exc_class=exceptions.CrdsError):
                after_context = self.resolve_context(self.args.after_context[0])
                after_ix = pmaps.index(after_context)
                contexts = pmaps[after_ix:]
        elif config.get_crds_env_context():
            contexts = [self.resolve_context(config.get_crds_env_context())]
        else:
            contexts = [self.resolve_context(self.observatory + "-operational")]
        log.verbose("Determined contexts: ", contexts, verbosity=55)
        return sorted(contexts)

    def _list_mappings(self, glob_pattern="*.pmap"):
        """Return a list of all the .pmap's on the CRDS Server."""
        self.require_server_connection()
        return heavy_client.list_mappings(self.observatory, glob_pattern)

    def _list_references(self, glob_pattern="*"):
        self.require_server_connection()
        return api.list_references(self.observatory, glob_pattern)

    def _orphan_references(self):
        """Return the set of reference files which exist the CRDS catalog but
        which cannot be found within the reference files referred to by any .pmap.
        """
        all_references = self._list_references()
        all_references = set(all_references)
        all_pmaps = self._list_mappings("*.pmap")
        pmap_references = self._get_context_references(all_pmaps)
        return set(all_references - pmap_references)

    def dump_files(self, context, files=None, ignore_cache=None):
        """Download mapping or reference `files1` with respect to `context`,  tracking stats."""
        if ignore_cache is None:
            ignore_cache = self.args.ignore_cache
        _localpaths, downloads, nbytes = api.dump_files(
            context, files, ignore_cache=ignore_cache, raise_exceptions=self.args.pdb)
        self.increment_stat("total-files", downloads)
        self.increment_stat("total-bytes", nbytes)

    def get_context_mappings(self):
        """Return the list of mappings which are pointed to by the mappings in `self.contexts`.

        If --all was specified,  include orphan mappings not reachable from any .pmap.
        """
        if not self.contexts:
            return []
        log.verbose("Getting all mappings for specified contexts.", verbosity=55)
        if self.args.all:
            mappings = self._list_mappings("*.*map")
            with log.verbose_warning_on_exception("Downloading all mappings failed"):
                self.dump_files(self.default_context, mappings)
        else:
            mappings = self.contexts

        useable_contexts, mapping_closure = self._dump_and_vet_mappings(mappings)
        self.contexts = useable_contexts   # XXXX reset self.contexts
        log.verbose("Got mappings from specified (usable) contexts: ", self.contexts, verbosity=55)

        return mapping_closure

    def _dump_and_vet_mappings(self, mappings):
        """Ensure all mapping files listed in `mappings` are in the CRDS cache.

        Returns  (list(useable_contexts), list(mappings_closures))
        """
        log.verbose("Getting and checking specified mappings.", verbosity=55)

        # Based on the specified mappings,  identify the  mappings they refer to
        mapping_closure = set()
        for mapping in mappings:
            with log.warn_on_exception("Failed loading context", repr(mapping)):
                try:
                    loadable = rmap.get_cached_mapping(mapping)
                    loadable_names = loadable.mapping_names()
                except Exception as exc:
                    log.verbose("Failed loading", repr(mapping),
                                "using API call to get mapping names", str(exc))
                    loadable_names = api.get_mapping_names(mapping)
                mapping_closure |= set(loadable_names)

        # Dump all missing files in one call
        with log.verbose_warning_on_exception("Mapping closure download failed"):
            self.dump_files(self.default_context, mapping_closure)

        # After attempting to sync,  identify which mappings are actually loadable now.
        useable_contexts = []
        for mapping in mappings:
            with log.warn_on_exception("Failed loading", repr(mapping)):
                _loaded = rmap.get_cached_mapping(mapping)
                useable_contexts.append(mapping)

        return sorted(useable_contexts), sorted(mapping_closure)

    def _get_context_references(self, mappings=None):
        """Return the set of references which are pointed to by the mappings in `mappings`
        or self.get_context_mappings() when mappings is None.
        """
        references = set()
        mappings = self.get_context_mappings() if mappings is None else mappings
        for context in mappings:
            try:
                pmap = rmap.get_cached_mapping(context)
                references |= set(pmap.reference_names())
                log.verbose("Determined references from cached mapping", repr(context))
            except Exception:  # only ask the server if loading context fails
                log.verbose("Determined references from CRDS server service", repr(context))
                references |= set(api.get_reference_names(context))
        return references

    def get_context_references(self):
        """Return the sorted list of references defined by the closure of the specified
        contexts.   If --include-orphans was specified,  also include reference files
        from the CRDS catalog which do not appear in any .pmap.
        """
        references = self._get_context_references(self.contexts)
        if self.args.include_orphans:
            references |= self._orphan_references()
        return sorted(list(references))

def expand_all_instruments(observatory, context):
    """Expand symbolic context specifiers for rmaps with "all" for instrument
    into the list of rmaps for every instrument in the related context (e.g. edit or operational).

    e.g.  jwst-all-photom-operational -->  [jwst-miri-photom-operational, jwst-nircam-photom-operational, ...]

    Expansion of "all" is determined by instruments in e.g. jwst-operational
    """
    mtch = config.CONTEXT_RE.match(context)
    if mtch and mtch.group("instrument") == "all":
        root_context = observatory + "-" + mtch.group("date")
        pmap = heavy_client.get_symbolic_mapping(root_context)
        observatory = mtch.group("observatory")
        filekind = mtch.group("filekind")
        date = mtch.group("date")
        all_maps = [ "-".join([observatory, instrument, filekind, date])
                for instrument in pmap.selections.keys() if instrument != "system"]
    else:
        all_maps = [context]
    return all_maps



"""This module is a command line script which handles comparing the best
reference recommendations for a particular context and dataset files.

For more details on the several modes of operations and command line parameters browse the source or run:

% crds bestrefs --help
"""
import json
import gc

# ===================================================================

import crds
from crds.core import log, utils, heavy_client
from crds.core.exceptions import CrdsError
from crds import data_file, matches
from crds.client import api

import pickle

# ===================================================================

MIN_DATE = "1900-01-01 00:00:00"
MAX_DATE = "9999-01-01 23:59:59"

# ===================================================================
# There's a problem with using CDBS as a gold standard in getting consistent results between
# DADSOPS DB (fast) and running command line OPUS bestrefs (slow but definitive).   This is kludged
# around here with a two tiered scheme for getting dataset headers:  first load headers from a primary
# source:  files or DADSOPS.  Next update headers from a pickle file computed "elsewhere".   Elsewhere
# is a script that ssh'es to a DMS machine for each dataset and runs OPUS bestrefs,  then eventually
# saves a pickle.
# It is also possible to run solely off of a pickle file(s) (which decouples from the database,  useful for
# both consistency/control and to take a load off the database.)


class HeaderGenerator:
    """Generic source for lookup parameters and historical comparison results."""

    def __init__(self, context, sources, datasets_since):
        self.context = context
        self.observatory = None if context is None else utils.file_to_observatory(context)
        self.sources = sources
        self.headers = {}
        self._datasets_since = datasets_since

    def __iter__(self):
        """Return the sources from self with EXPTIME >= self.datasets_since."""
        for source in sorted(self.sources):
            with log.error_on_exception("Failed loading source", repr(source),
                                        "from", repr(self.__class__.__name__)):
                instrument = utils.header_to_instrument(self.header(source))
                exptime = matches.get_exptime(self.header(source))
                since = self.datasets_since(instrument)
                # since == None when no command line argument given.
                if since is None or exptime >= since:
                    yield source
                else:
                    log.verbose("Dropping source", repr(source),
                                "with EXPTIME =", repr(exptime),
                                "< --datasets-since =", repr(since))

    def datasets_since(self, instrument):
        """Return the earliest dataset processed cut-off date for `instrument`.

        If a universal since-date is in effect,  just return it.

        If the since-date varies by instrument, but is not defined for `instrument`, return
        a value equivalent to the end-of-time since it means that no references for `instrument`
        were identified by --datsets-since=auto.
        """
        if isinstance(self._datasets_since, dict):
            return self._datasets_since.get(instrument.lower(),  MIN_DATE)
        else:
            return self._datasets_since

    def header(self, source):
        """Return the full header corresponding to `source`.   If header is a string, raise an exception."""
        header = self._header(source)
        if isinstance(header, str):
            raise CrdsError("Failed to fetch header for " + repr(source) + ": " + repr(header))
        else:
            return dict(header)

    def _header(self, source):
        """Return the full header corresponding to `source`.   Source is a dataset id or filename."""
        return self.headers[source]

    def get_lookup_parameters(self, source):
        """Return the parameters corresponding to `source` used to drive a best references lookup."""
        return add_instrument(self.header(source))

    def get_old_bestrefs(self, source):
        """Return the historical best references corresponding to `source`.  Always define old bestrefs
        in terms of filekind/typename rather than in terms of FITS keyword.
        """
        header = add_instrument(self.header(source))
        instrument = utils.header_to_instrument(header)
        pmap = crds.get_pickled_mapping(self.context)
        result = {}
        for filekind in pmap.get_imap(instrument).selections:
            keyword = pmap.locate.filekind_to_keyword(filekind)
            filekind = filekind.upper()
            try:
                result[filekind] = header[keyword]
            except KeyError:
                result[filekind] = header.get(filekind, "UNDEFINED")
        return result

    def save_pickle(self, outpath, only_ids=None):
        """Write out headers to `outpath` file which can be a Python pickle or .json"""
        if only_ids is None:
            only_hdrs = self.headers
        else:
            only_hdrs = {dataset_id: hdr for (dataset_id, hdr) in self.headers.items() if dataset_id in only_ids}
        log.info("Writing all headers to", repr(outpath))
        if outpath.endswith(".json"):
            with open(outpath, "w+") as pick:
                for dataset, header in sorted(only_hdrs.items()):
                    pick.write(json.dumps({dataset: header}) + "\n")
        elif outpath.endswith(".pkl"):
            with open(outpath, "wb+") as pick:
                pickle.dump(only_hdrs, pick)
        log.info("Done writing", repr(outpath))

    def update_headers(self, headers2, only_ids=None):
        """Incorporate `headers2` updated values into `self.headers`.  Since `headers2` may be incomplete,
        do param-by-param update.   Nominally,  this is to add OPUS bestrefs corrections (definitive) to DADSOPS
        database bestrefs (fast but not definitive).
        """
        if only_ids is None:
            only_ids = headers2.keys()

        items = headers2.items()
        for dataset_id, header in items:
            if isinstance(header, str):
                log.warning("Skipping bad dataset", dataset_id, ":", headers2[dataset_id])
                del headers2[dataset_id]

        # Munge for consistent case and value formatting regardless of source
        headers2 = {dataset_id:
                    {key.upper(): bestrefs_condition(val) for (key, val) in headers2[dataset_id].items()}
                    for dataset_id in headers2 if dataset_id in only_ids}

        # replace param-by-param,  not id-by-id, since headers2[id] may be partial
        for dataset_id in headers2:
            if dataset_id not in self.headers:
                self.headers[dataset_id] = {}
            header1, header2 = self.headers[dataset_id], headers2[dataset_id]
            for key in header2:
                if key not in header1 or header1[key] != header2[key]:
                    header1[key] = header2[key]

    def handle_updates(self, all_updates):
        """Base handle_updates() updates the loaded headers with the computed bestrefs for use
        with --save-pickle.
        """
        for dataset in sorted(all_updates):
            updates = all_updates[dataset]
            if updates:
                log.verbose("-" * 120)
                for update in sorted(updates):
                    new_ref = update.new_reference.upper()
                    if new_ref != "N/A":
                        new_ref = new_ref.lower()
                    self.headers[dataset][update.filekind.upper()] = new_ref


def bestrefs_condition(value):
    """Condition header keyword value to normal form,  converting NOT FOUND N/A to N/A."""
    val = utils.condition_value(value)
    if val == "NOT FOUND N/A":
        val = "N/A"
    return val

# ===================================================================

# FileHeaderGenerator uses a deferred header loading scheme which incrementally reads each header
# from a file as processing is going on via header().   The "pickle correction" scheme works by
# pre-loading the FileHeaderGenerator with pickled headers...  which prevents the file from ever being
# accessed (except possibly to update the headers).


class FileHeaderGenerator(HeaderGenerator):
    """Generates lookup parameters and old bestrefs from dataset files."""

    def _header(self, filename):
        """Get the best references recommendations recorded in the header of file `dataset`."""
        gc.collect()
        if filename not in self.headers:
            self.headers[filename] = data_file.get_free_header(filename, (), None, self.observatory)
        return self.headers[filename]

    def handle_updates(self, all_updates):
        """Write best reference updates back to dataset file headers."""
        super(FileHeaderGenerator, self).handle_updates(all_updates)
        for source in sorted(all_updates):
            updates = all_updates[source]
            if updates:
                log.verbose("-" * 120)
                update_file_bestrefs(self.context, source, updates)

# ===================================================================


class DatasetHeaderGenerator(HeaderGenerator):
    """Generates lookup parameters and historical best references from dataset ids.   Server/DB bases"""

    def __init__(self, context, datasets, datasets_since):
        """"Contact the CRDS server and get headers for the list of `datasets` ids with respect to `context`."""
        super(DatasetHeaderGenerator, self).__init__(context, datasets, datasets_since)
        server = api.get_crds_server()
        log.info("Dumping dataset parameters from CRDS server at", repr(server), "for", repr(datasets))
        self.headers = api.get_dataset_headers_by_id(context, datasets)
        log.info("Dumped", len(self.headers), "of", len(datasets), "datasets from CRDS server at", repr(server))

        # every command line id should correspond to 1 or more headers
        for source in self.sources:
            if self.matching_two_part_id(source) not in self.headers.keys():
                log.warning("Dataset", repr(source), "isn't represented by downloaded parameters.")

        # Process according to downloaded 2-part ids,  not command line ids.
        self.sources = sorted(self.headers.keys())

    def matching_two_part_id(self, source):
        """Convert any command line dataset id into it's matching two part id.

        matching_two_part_id(<association>)                  -->  <association>  : <first_member>

        matching_two_part_id(<association>:<member>)         -->  <association>  : <member>
        matching_two_part_id(<unassociated>)                 -->  <unassociated> : <unassociated>
        matching_two_part_id(<unassociated>:<unassociated>)  -->  <unassociated> : <unassociated>
        """
        parts = [_normalize_jwst_id_part(part) for part in source.split(":")]
        assert 1 <= len(parts) <= 2, "Invalid dataset id " + repr(source)
        try:    # when specifying datasets with 1-part id, return first of "associated ids"
                # when specifying datasets with 2-part id,
            if len(parts) == 1:
                return sorted(dataset_id for dataset_id in self.headers if parts[0] in dataset_id)[0]
            else:
                return source
        except Exception:
            return source

def _normalize_jwst_id_part(part):
    """Converts jw88600071001_02101_00001_nrs1  --> jw88600071001_02101_00001.nrs.   The former is
    common notation for most dataset usages,  the latter is the official form for the web API to
    the archive parameter service for JWST.
    """
    if "_" in part and "." not in part:  # not HST and common JWST parlance
        bits = part.split("_")
        fileSetName = "_".join(bits[:-1])
        detector = bits[-1]
        return fileSetName + "." + detector   # Formal archive API
    else:
        return part

class InstrumentHeaderGenerator(HeaderGenerator):
    """Generates lookup parameters and historical best references from a list of instrument names.  Server/DB based."""

    def __init__(self, context, instruments, datasets_since, save_pickles, server_info):
        """"Contact the CRDS server and get headers for the list of `instruments` names with respect to `context`."""
        super(InstrumentHeaderGenerator, self).__init__(context, [], datasets_since)
        self.instruments = instruments
        self.sources = self.determine_source_ids()
        self.save_pickles = save_pickles
        try:
            self.segment_size = server_info.max_headers_per_rpc
        except Exception:
            self.segment_size = 5000

    def determine_source_ids(self):
        """Return the dataset ids for all instruments."""
        server = api.get_crds_server()
        source_ids = []
        for instrument in self.instruments:
            since_date = self.datasets_since(instrument)
            if since_date:
                log.info("Dumping dataset parameters for", repr(instrument), "from CRDS server at", repr(server),
                         "since", repr(since_date))
            else:
                log.info("Dumping dataset parameters for", repr(instrument), "from CRDS server at", repr(server))
            instr_ids = api.get_dataset_ids(self.context, instrument, self.datasets_since(instrument))
            log.info("Downloaded ", len(instr_ids), "dataset ids for", repr(instrument), "since", repr(since_date))
            source_ids.extend(instr_ids)
        return sorted(source_ids)  # sort is needed to match generic __iter__() sort. assumes instruments don't shuffle

    def _header(self, source):
        """Return the header associated with dataset id `source`,  fetching the surround segment of
        headers if `source` is not already in the cached set of headers.
        """
        if source not in self.headers:
            self.fetch_source_segment(source)
        return self.headers[source]

    def fetch_source_segment(self, source):
        """Return the segment of dataset ids which surrounds id `source`."""
        try:
            index = self.sources.index(source) // self.segment_size
        except ValueError as exc:
            raise CrdsError("Unknown dataset id " + repr(source)) from exc
        lower = index * self.segment_size
        upper = (index + 1) * self.segment_size
        segment_ids = self.sources[lower:upper]
        log.verbose("Dumping", len(segment_ids), "datasets from indices", lower, "to",
                    lower + len(segment_ids), verbosity=20)
        dumped_headers = api.get_dataset_headers_by_id(self.context, segment_ids)
        log.verbose("Dumped", len(dumped_headers), "datasets", verbosity=20)
        if self.save_pickles:  # keep all headers,  causes memory problems with multiple instruments on ~8G ram.
            self.headers.update(dumped_headers)
        else:  # conserve memory by keeping only the last N headers
            self.headers = dumped_headers


class PickleHeaderGenerator(HeaderGenerator):
    """Generates lookup parameters and historical best references from a list of pickle files (or .json files)
    using successive updates to sets of header dictionaries.  Trailing pickles override leading pickles.
    """

    def __init__(self, context, pickles, datasets_since, only_ids=None):
        """"Contact the CRDS server and get headers for the list of `datasets` ids with respect to `context`."""
        super(PickleHeaderGenerator, self).__init__(context, pickles, datasets_since)
        for pickle in pickles:
            log.info("Loading file", repr(pickle))
            pick_headers = load_bestrefs_headers(pickle)
            if not self.headers:
                log.info("Loaded", len(pick_headers), "datasets from file", repr(pickle),
                         "completely replacing existing headers.")
                self.headers.update(pick_headers)   # replace all of dataset_id
            else:  # OPUS bestrefs don't include original matching parameters,  so full replacement doesn't work.
                log.info("Loaded", len(pick_headers), "datasets from file", repr(pickle),
                         "augmenting existing headers.")
                self.update_headers(pick_headers, only_ids=only_ids)
        self.sources = only_ids or self.headers.keys()

# ============================================================================

def load_bestrefs_headers(path):
    """Given `path` to a serialization file,  load  {dataset_id : header, ...}.
    Supports .pkl and .json.

    For easier editing and syntax error precision,  .json files are stored as
    one header per line.

    Also used by server to load mock parameters.
    """
    if path.endswith(".json"):
        headers = {}
        try:
            with open(path, "r") as pick:
                for line in pick:
                    if line.strip():
                        headers.update(json.loads(line))
        except ValueError:
            with open(path, "r") as pick:
                headers = json.load(pick)
    elif path.endswith(".pkl"):
        with open(path, "rb") as pick:
            headers = pickle.load(pick)
    else:
        raise ValueError("Valid serialization formats are .json and .pkl")
    return headers

def add_instrument(header):
    """Add INSTRUME keyword."""
    instrument = utils.header_to_instrument(header)
    header["INSTRUME"] = instrument
    header["META.INSTRUMENT.NAME"] = instrument
    return header

def update_file_bestrefs(context, dataset, updates):
    """Update the header of `dataset` with best reference recommendations
    `bestrefs` determined by context named `pmap`.
    """
    if not updates:
        return

    version_info = heavy_client.version_info()
    instrument = updates[0].instrument
    locator = utils.instrument_to_locator(instrument)
    prefix = locator.get_env_prefix(instrument)

    with data_file.fits_open(dataset, mode="update", do_not_scale_image_data=True, checksum=False) as hdulist:

        def set_key(keyword, value):
            """Set a single keyword value with logging,  bound to outer-scope hdulist."""
            log.verbose("Setting", repr(dataset), keyword, "=", value)
            hdulist[0].header[keyword] = value

        set_key("CRDS_CTX", context)
        set_key("CRDS_VER", version_info)

        for update in sorted(updates):
            new_ref = update.new_reference.upper()
            if new_ref != "N/A":
                new_ref = (prefix + new_ref).lower()
            keyword = locator.filekind_to_keyword(update.filekind)
            set_key(keyword, new_ref)

        # This is a workaround for a bug in astropy.io.fits handling of
        # FITS updates that are header-only and extend the header.
        # This statement appears to do nothing but *is not* pointless.
        for hdu in hdulist:
            hdu.data



"""This module supports conditional substitutions on reference file header parameters
at reference file submission time.  It's principle function is to expand wild cards into
explicit matching patterns in the CRDS rmaps.

CRDS defines a substitutions file with a dictionary of this form:

    { instrument : { header_key :  { condition_expr: expansion } } }

When a reference file is submitted,  the required parameters are located in the substitutions
dictionary, if applicable.   For each `header_key`,  `condition_expr` expressions are evaluated
with respect to the reference file header.  If the `condition_expr` is True,  the value of the
`header_key` is literally replaced with `expansion` for the purpose of rmap matching.  The file
is not modified,  the rmap reinterprets the file.

An example substitution rule is:

    { "CCDGAIN" : { "DETECTOR='WFC3' and CCDGAIN='-999'" : "1.0|2.0|4.0|8.0" }

CRDS defines a function:

   expanded_header = expand_wildcards(instrument, header)

which will interpret the rules to expand appropriate values in header.
"""
import os.path

from . import log, utils, selectors
# from . import heavy_client (circular for rmap)

# ============================================================================

HERE = os.path.dirname(__file__) or "."

# ============================================================================

class HeaderExpander:
    """HeaderExpander applies a set of expansion rules to a header.  It
    compiles the applicability expression of each rule.

    >>> from pprint import pprint as pp

    >>> expansions = {
    ...  'FILTER1' : { "DETECTOR=='HRC' and FILTER1=='ANY'": 'F555W|F775W|F625W'},
    ...  'OBSTYPE' : { "DETECTOR=='HRC' and FILTER1=='G800L' and OBSTYPE=='ANY'": 'IMAGING|CORONAGRAPHIC'},
    ... }
    >>> expander = HeaderExpander(expansions)

    >>> header = { "DETECTOR":"HRC", "FILTER1":"ANY" }
    >>> pp(expander.expand(header))
    {'DETECTOR': 'HRC', 'FILTER1': 'F555W|F775W|F625W'}

    >>> header = { "DETECTOR":"HRC", "FILTER1":"G280L", "OBSTYPE":"ANY" }
    >>> pp(expander.expand(header))
    {'DETECTOR': 'HRC', 'FILTER1': 'G280L', 'OBSTYPE': 'ANY'}

    >>> header = { "DETECTOR":"HRC", "FILTER1":"G800L", "OBSTYPE":"ANY" }
    >>> pp(expander.expand(header))
    {'DETECTOR': 'HRC', 'FILTER1': 'G800L', 'OBSTYPE': 'IMAGING|CORONAGRAPHIC'}
    """
    def __init__(self, expansion_mapping, expansion_file="(none)"):
        self.mapping = {}
        for var, substitutes in expansion_mapping.items():
            for expr, replacement in substitutes.items():
                self.mapping[(var, expr)] = (replacement, compile(expr, expansion_file, "eval"))  # compiled code is from static file.
        self._required_keys = self.required_keys()

    def expand(self, header, required_parkeys=None):
        """Given a reference matching `header`,  evaluate all the expansion rules with respect
        to it and return a modified copy which include substitutions.
        """
        header = dict(header)
        expanded = dict(header)
        if required_parkeys is None:
            required_parkeys = header.keys()
        log.verbose("Unexpanded header", self.required_header(header))
        for (var, expr), (expansion, compiled) in self.mapping.items():
            if var not in required_parkeys:
                log.verbose("Skipping expansion for unused parkey", repr(var),
                            "of", repr(expansion))
                continue
            try:
                applicable = eval(compiled, {}, header)  # compiled code is from static file.
            except Exception as exc:
                log.verbose_warning("Header expansion for",repr(expr),
                            "failed for", repr(str(exc)))
                continue
            if applicable:
                log.verbose("Expanding", repr(expr), "yields",
                            var + "=" + repr(expansion))
                expanded[var] = expansion
            else:
                log.verbose("Expanding", repr(expr), "doesn't apply.")
        log.verbose("Expanded header", self.required_header(expanded))
        return expanded

    def required_keys(self):
        """Return the list of header keys required to evaluate all substitutions conditions."""
        required = []
        for (_var, expr) in self.mapping:
            required.extend(required_keys(expr))
        return sorted(set(required))

    def required_header(self, header):
        """Ensure all required keywords for evaluating expansions are defined in `header`, at lest
        with a value of UNDEFINED.
        """
        return sorted({ key: header.get(key, "UNDEFINED") for key in self._required_keys }.items())

    def get_expansion_values(self):
        values = {}
        for (var, expr), (expansion, compiled) in self.mapping.items():
            if var not in values:
                values[var] = set()
            values[var] |= set([value for value in selectors.glob_list(expansion)
                                if not value.startswith("BETWEEN")])
        return { var : sorted(list(vals)) for (var, vals) in values.items() }

def required_keys(expr):
    """
    >>> required_keys("VAR1=='VAL1' and VAR2=='VAL2' and VAR3=='VAL3'")
    ['VAR1', 'VAR2', 'VAR3']
    """
    return sorted({term.split("=")[0].strip() for term in expr.split("and")})


class ReferenceHeaderExpanders(dict):
    """Container class for all expanders for all instruments of an observatory."""

    @classmethod
    @utils.cached
    def load(cls, observatory):
        """Load the substution rules from the observatory package directory."""
        pkg = utils.get_observatory_package(observatory)
        rules = utils.evalfile(pkg.HERE + "/substitutions.dat")
        expanders = ReferenceHeaderExpanders()
        for instrument in rules:
            expanders[instrument] = HeaderExpander(rules[instrument])
        return expanders

    def expand_wildcards(self, rmapping, header):
        """Transform header values according to expansion rules."""
        try:
            header = self[rmapping.instrument].expand(header, rmapping.get_required_parkeys())
        except KeyError:
            header = dict(header)
        # log.warning("Unknown instrument", repr(instrument), " in expand_wildcards().")
        return header

    def validate_expansions(self, pmap):
        for instrument in sorted(pmap.selections):
            if instrument not in self:
                log.verbose("Instrument", repr(instrument), "has no substitutions.")
                continue
            imap = pmap.get_imap(instrument)
            valid_values = imap.get_valid_values_map(condition=True, remove_special=False)
            for parameter, values in sorted(self[instrument].get_expansion_values().items()):
                for value in sorted(values):
                    if parameter not in valid_values or not valid_values[parameter]:
                        log.verbose("For", repr(instrument), "parameter", repr(parameter),
                                    "with value", repr(value), "is unchecked.")
                        continue
                    if value not in valid_values[parameter]:
                        log.error("For", repr(instrument), "parameter", repr(parameter), "value",
                                  repr(value), "is not in", valid_values[parameter])

def expand_wildcards(rmapping, header):
    """Expand substitution values in `header` with respect to the instrument and observatory
    defined in `rmapping`.
    """
    expanders = ReferenceHeaderExpanders.load(rmapping.observatory)
    return expanders.expand_wildcards(rmapping, header)


def validate_substitutions(pmap_name):
    """This is a unit test function."""
    from . import heavy_client
    pmap = heavy_client.get_symbolic_mapping(pmap_name)
    expanders = ReferenceHeaderExpanders.load(pmap.observatory)
    expanders.validate_expansions(pmap)



'''
Created on Feb 15, 2017

@author: jmiller
'''
import re

# ============================================================================

import contextlib
import uuid
import os
import io

from astropy.io import fits

import numpy as np

# ============================================================================

from crds.core import config, utils, log

from .abstract import AbstractFile, hijack_warnings

# ============================================================================

@hijack_warnings
def fits_open_trapped(filename, **keys):
    """Same as fits_open but with some astropy and JWST DM warnings hijacked by CRDS.
    This is predominantly an interface for crds.certify and nominally verifies
    checksums although technically they are driven by CRDS_FITS_VERIFY_CHECKSUM.
    """
    return fits_open(filename, **keys)

@contextlib.contextmanager
@utils.gc_collected
def fits_open(filename, **keys):
    """Return the results of io.fits.open() configured using CRDS environment
    settings, overridden by any conflicting keyword parameter values.
    Nominally used for updating bestrefs FITS headers.

    CRDS_FITS_VERIFY_CHECKSUM is used to enable/disable default checksum verification.
    CRDS_FITS_IGNORE_MISSING_END is used to enable/disable the missing FITS END check.
    """
    keys = dict(keys)
    if "checksum" not in keys:
        keys["checksum"] = bool(config.FITS_VERIFY_CHECKSUM)
    if "ignore_missing_end" not in keys:
        keys["ignore_missing_end"] = bool(config.FITS_IGNORE_MISSING_END)
    handle = None
    try:
        handle = fits.open(filename, **keys)
        yield handle
    finally:
        if handle is not None:
            handle.close()

def get_fits_header_union(filepath, needed_keys=(), original_name=None, observatory=None, **keys):
    """Get the union of keywords from all header extensions of FITS
    file `fname`.  In the case of collisions, keep the first value
    found as extensions are loaded in numerical order.

    DOES NOT verify FITS checksums.
    """
    file_obj = FitsFile(filepath)
    header = file_obj.get_header(needed_keys, checksum=False)
    log.verbose("Header of", repr(filepath), "=", log.PP(header), verbosity=90)
    return header

# ============================================================================

class FitsFile(AbstractFile):

    format = "FITS"

    def get_info(self):
        """Capture the output from the fits info() function."""
        s = io.StringIO()
        fits.info(self.filepath, s)
        s.seek(0)
        info_string = "\n".join(s.read().splitlines()[1:])
        return info_string

    def _array_name_to_hdu_index(self, array_name):
        """Convert array names with extended notations into "index" values
        which can be used to select particular HDUs.

        The extended notations include specifying HDUs using an extension number
        pseudonym like this:

        # does next to nothing.
        >>> import os.path
        >>> path = os.path.abspath("tests/data/y951738kl_hv.fits")
        >>> fits_file = FitsFile(path)

         # doesn't require a real file or validate that it exists
        >>> fits_file._array_name_to_hdu_index("EXTENSION5")
        ('5', 5)

         # doesn't require a real file or validate that it exists
        >>> fits_file._array_name_to_hdu_index("EXT5")
        ('5', 5)

        or by name but referring to a particular "ver" like this:

        # info about file structure, real file required for this usecase
        finfo crds/tests/data/y951738kl_hv.fits
        -----------------------------------------------------
        Filename: crds/tests/data/y951738kl_hv.fits
        No.    Name      Ver    Type      Cards   Dimensions   Format
        0  PRIMARY       1 PrimaryHDU      22   ()
        1  FUVA          1 BinTableHDU     17   106R x 2C   [D, I]
        2  FUVB          1 BinTableHDU     17   127R x 2C   [D, I]

        # Actual test and expected results
        >>> fits_file._array_name_to_hdu_index("FUVB")
        ('FUVB', 2)

        """
        array_name = str(array_name)
        ext_match = re.match(r"(EXT(ENSION)?)?(\d+)", array_name)
        if ext_match:
            i = ext_match.group(3)
            return i, int(i)
        ver_match = re.match(r"(.*)__(\d+)", array_name)
        if ver_match:
            name, ver = ver_match.group(1), int(ver_match.group(2))
            return (name, ver), self._extension_number((name, ver))
        return array_name, self._extension_number(array_name)

    def _extension_number(self, index):
        """Converts an HDU `index` value returned by _array_name_to_hdu_index() or HD
        name (implicit ver 1) into a FITS HDU number.
        """
        with fits_open(self.filepath) as hdus:
            for i, hdu in enumerate(hdus):
                if isinstance(index, str):
                    if hdu.name == index:
                        return i
                elif isinstance(index, tuple):
                    if hdu.name == index[0] and hdu.ver == index[1]:
                        return i
                else:
                    raise ValueError(
                        "Unrecognized HDU index format: " + str(index))
            raise ValueError("Can't find HDU: " + str(index))

    def get_array(self, array_name):
        """Return the `name`d array data from `filepath`,  alternately indexed
        by `extension` number.
        """
        index = self._array_name_to_hdu_index(array_name)
        with fits_open(self.filepath) as hdus:
            return hdus[index[1]].data

    def get_raw_header(self, needed_keys=(), **keys):
        """Get the union of keywords from all header extensions of FITS
        file `fname`.  In the case of collisions, keep the first value
        found as extensions are loaded in numerical order.
        """
        union = []
        with fits_open(self.filepath, **keys) as hdulist:
            for hdu in hdulist:
                for card in hdu.header.cards:
                    card.verify('fix')
                    key, value = card.keyword, str(card.value)
                    if not key:
                        continue
                    union.append((key, value))
        return union

    def get_array_properties(self, array_name, keytype="A"):
        """Return a Struct defining the properties of the FITS array in extension named `array_name`."""
        with fits_open(self.filepath) as hdulist:
            try:
                array_name = self._array_name_to_hdu_index(array_name)
                hdu = hdulist[array_name[1]]
            except Exception:
                return 'UNDEFINED'
            generic_class = {
                "IMAGEHDU" : "IMAGE",
                "BINTABLEHDU" : "TABLE",
            }.get(hdu.__class__.__name__.upper(), "UNKNOWN")
            if generic_class in ["IMAGE","UNKNOWN"]:
                typespec = hdu.data.dtype.name
                column_names = None
            else: # TABLE
                dtype = hdu.data.dtype
                typespec = {name.upper():str(dtype.fields[name][0]) for name in dtype.names}
                column_names = [name.upper() for name in hdu.data.dtype.names]
            return utils.Struct(
                        SHAPE = hdu.data.shape,
                        KIND = generic_class,
                        DATA_TYPE = typespec,
                        COLUMN_NAMES = column_names,
                        NAME = array_name[0],
                        EXTENSION = array_name[1],
                        DATA = hdu.data if (keytype == "D") else None
                    )

    # ----------------------------------------------------------------------------------------------

    def getval(self, key):
        """FITS version of getval() with checksum=False added."""
        return super(FitsFile, self).getval(key, checksum=False)

    def setval(self, key, value):
        """FITS version of setval() method."""
        fits.setval(self.filepath, key, value=value)

    @hijack_warnings
    def add_checksum(self):
        """Add checksums to `filepath`."""
        output = "crds-" + str(uuid.uuid4()) + ".fits"
        with fits_open(self.filepath, do_not_scale_image_data=True) as hdus:
            for hdu in hdus:
                data = hdu.data if hdu.data is not None else np.array([])
                fits.append(output, data, hdu.header, checksum=True)
        os.remove(self.filepath)
        os.rename(output, self.filepath)

    @hijack_warnings
    def remove_checksum(self):
        """Remove checksums from `filepath`."""
        output = "crds-" + str(uuid.uuid4()) + ".fits"
        with fits_open(self.filepath, checksum=False, do_not_scale_image_data=True) as hdus:
            for hdu in hdus:
                hdu.header.pop("CHECKSUM",None)
                hdu.header.pop("DATASUM", None)
                data = hdu.data if hdu.data is not None else np.array([])
                fits.append(output, data, hdu.header, checksum=False)
        os.remove(self.filepath)
        os.rename(output, self.filepath)

    @hijack_warnings
    def verify_checksum(self):
        """Verify checksums in `filepath`."""
        with fits.open(self.filepath, do_not_scale_image_data=True, checksum=True) as hdus:
            for hdu in hdus:
                hdu.verify("warn")

    def get_asdf_standard_version(self):
        """
        Return the ASDF Standard version associated with this file as a string,
        or `None` if the file is neither an ASDF file nor contains an embedded
        ASDF file.
        """
        with fits.open(self.filepath) as hdus:
            asdf_hdu = next((hdu for hdu in hdus if hdu.name == "ASDF"), None)
            if asdf_hdu is None:
                return None
            else:
                import asdf
                buff = io.BytesIO(asdf_hdu.data)
                with asdf.open(buff) as handle:
                    return str(handle.version)

def test():
    from crds.io import fits
    import doctest
    return doctest.testmod(fits)

if __name__ == "__main__":
    print(test())



"""This module provides functions which determine various observatory
specific policies for TOBS:

1. How to convert reference file basenames to fully specified paths.

2. How to manage parameters for reference file Validator objects used
in the certification of reference files.

"""
import os.path
import re

from crds.core import log, rmap, config, utils, timestamp
from crds import data_file
from crds.io import abstract

HERE = os.path.dirname(__file__) or "./"

# =======================================================================

# These two functions decouple the generic reference file certifier program
# from observatory-unique ways of specifying and caching Validator parameters.

from crds.tobs import TYPES, INSTRUMENTS, FILEKINDS, EXTENSIONS, INSTRUMENT_FIXERS, TYPE_FIXERS

get_row_keys_by_instrument = TYPES.get_row_keys_by_instrument
get_item = TYPES.get_item
suffix_to_filekind = TYPES.suffix_to_filekind
filekind_to_suffix = TYPES.filekind_to_suffix
get_all_tpninfos = TYPES.get_all_tpninfos

# =============================================================================
# Plugin-functions for this observatory,  accessed via locator.py

HERE = os.path.dirname(__file__) or "."

def tpn_path(tpn_file):
    return os.path.join(HERE, "tpns", tpn_file)

def get_extra_tpninfos(refpath):
    return []

def project_check(refpath):
    pass

def get_exptypes(instrument=None):
    """Return the list of EXP_TYPE values for instrument,  or for all
    instruments if instrument is not specified.
    """
    raise NotImplementedError("TOBS has not defined get_exptypes().")

# =======================================================================

# When loading headers,  make sure each keyword in a tuple is represented with
# the same value enabling any form to be used.
CROSS_STRAPPED_KEYWORDS = {
       # "META.INSTRUMENT.NAME" : ["INSTRUME", "INSTRUMENT", "META.INSTRUMENT.TYPE",],
    }

@utils.cached
def get_static_pairs():
    return abstract.equivalence_dict_to_pairs(CROSS_STRAPPED_KEYWORDS)

def get_cross_strapped_pairs(header):
    """Return the list of keyword pairs where each pair describes synonyms for the same
    piece of data.
    """
    return  get_static_pairs()

# =======================================================================

REF_EXT_RE = re.compile(r"\.fits|\.r\dh$")

def get_file_properties(filename):
    """Figure out (instrument, filekind, serial) based on `filename` which
    should be a mapping or FITS reference file.

    >> get_file_properties("./hst_acs_biasfile_0001.rmap")
    ('acs', 'biasfile')

    >> get_file_properties("./hst_acs_biasfile_0001.pmap")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid .pmap filename './hst_acs_biasfile_0001.pmap'

    >> get_file_properties("test_data/s7g1700gl_dead.fits")
    """
    if config.is_mapping(filename):
        return decompose_newstyle_name(filename)[2:4]
    elif REF_EXT_RE.search(filename):
        result = get_reference_properties(filename)[2:4]
    else:
        try:
            result = properties_inside_mapping(filename)
        except Exception as exc:
            result = get_reference_properties(filename)[2:4]
    assert result[0] in INSTRUMENTS+[""], "Bad instrument " + \
        repr(result[0]) + " in filename " + repr(filename)
    assert result[1] in FILEKINDS+[""], "Bad filekind " + \
        repr(result[1]) + " in filename " + repr(filename)
    return result

def decompose_newstyle_name(filename):
    """
    >> decompose_newstyle_name("./hst.pmap")
    ('.', 'hst', '', '', '', '.pmap')

    >> decompose_newstyle_name("./hst_0001.pmap")
    ('.', 'hst', '', '', '0001', '.pmap')

    >> decompose_newstyle_name("./hst_acs.imap")
    ('.', 'hst', 'acs', '', '', '.imap')

    >> decompose_newstyle_name("./hst_acs_0001.imap")
    ('.', 'hst', 'acs', '', '0001', '.imap')

    >> decompose_newstyle_name("./hst_acs_biasfile.rmap")
    ('.', 'hst', 'acs', 'biasfile', '', '.rmap')

    >> decompose_newstyle_name("./hst_acs_biasfile_0001.rmap")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.rmap')

    >> decompose_newstyle_name("./hst_acs_biasfile.fits")
    ('.', 'hst', 'acs', 'biasfile', '', '.fits')

    >> decompose_newstyle_name("./hst_acs_biasfile_0001.fits")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.fits')
    """
    path, parts, ext = _get_fields(filename)
    observatory = parts[0]
    serial = list_get(parts, 3, "")

    if ext == ".pmap":
        assert len(parts) in [1,2], "Invalid .pmap filename " + repr(filename)
        instrument, filekind = "", ""
        serial = list_get(parts, 1, "")
    elif ext == ".imap":
        assert len(parts) in [2,3], "Invalid .imap filename " + repr(filename)
        instrument = parts[1]
        filekind = ""
        serial = list_get(parts, 2, "")
    else:
        assert len(parts) in [3,4], "Invalid filename " + repr(filename)
        instrument = parts[1]
        filekind = parts[2]
        serial = list_get(parts, 3, "")

    assert instrument in INSTRUMENTS+[""], "Invalid instrument " + \
        repr(instrument) + " in name " + repr(filename)
    assert filekind in FILEKINDS+[""], "Invalid filekind " + \
        repr(filekind) + " in name " + repr(filename)
    assert re.fullmatch(r"\d*", serial), "Invalid id field " + \
        repr(id) + " in name " + repr(filename)
    # extension may vary for upload temporary files.

    return path, observatory, instrument, filekind, serial, ext

def properties_inside_mapping(filename):
    """Load `filename`s mapping header to discover and
    return (instrument, filekind).
    """
    mapping = rmap.fetch_mapping(filename)
    if mapping.filekind == "PIPELINE":
        result = "", ""
    elif mapping.filekind == "INSTRUMENT":
        result = mapping.instrument, ""
    else:
        result = mapping.instrument, mapping.filekind
    return result

def _get_fields(filename):
    path = os.path.dirname(filename)
    name = os.path.basename(filename)
    name, ext = os.path.splitext(name)
    parts = name.split("_")
    return path, parts, ext

def list_get(l, index, default):
    try:
        return l[index]
    except IndexError:
        return default

CDBS_DIRS_TO_INSTR = {
   "/jref/":"acs",
   "/oref/":"stis",
   "/iref/":"wfc3",
   "/lref/":"cos",
   "/nref/":"nicmos",

   "/upsf/":"wfpc2",
   "/uref/":"wfpc2",
   "/uref_linux/":"wfpc2",

   "/yref/" : "fos",
   "/zref/" : "hrs",

}

def get_reference_properties(filename):
    """Figure out FITS (instrument, filekind, serial) based on `filename`.
    """
    try:   # Hopefully it's a nice new standard filename, easy
        return decompose_newstyle_name(filename)
    except AssertionError:  # cryptic legacy paths & names, i.e. reality
        pass
    # If not, dig inside the FITS file, slow
    return ref_properties_from_header(filename)

def ref_properties_from_header(filename):
    """Look inside FITS `filename` header to determine instrument, filekind.
    """
    # For legacy files,  just use the root filename as the unique id
    path, _parts, ext = _get_fields(filename)
    serial = os.path.basename(os.path.splitext(filename)[0])
    header = data_file.get_header(filename)
    instrument = header["INSTRUME"].lower()
    filetype = header["FILETYPE"].lower()
    filekind = TYPES.filetype_to_filekind(instrument, filetype)
    return path, "tobs", instrument, filekind, serial, ext



# =======================================================================

def header_to_reftypes(header, context):
    """Based on `header` return the default list of appropriate reference type names."""
    return [] # translates to all types.

# =======================================================================

def reference_keys_to_dataset_keys(rmapping, header):
    """Given a header dictionary for a reference file,  map the header back to
    keys relevant to datasets.
    """
    result = dict(header)
    if "USEAFTER" in header:  # and "DATE-OBS" not in header:
        reformatted = timestamp.reformat_date(header["USEAFTER"]).split()
        result["DATE-OBS"] = reformatted[0]
        result["TIME-OBS"] = reformatted[1]
    return result

# =======================================================================

def condition_matching_header(rmapping, header):
    """Condition the matching header values to the normalized form of the .rmap"""
    return utils.condition_header(header)

# =======================================================================

def get_env_prefix(instrument):
    """Return the environment variable prefix (IRAF prefix) for `instrument`."""
    return "crds://"

def filekind_to_keyword(filekind):
    """Return the FITS keyword at which a reference should be recorded."""
    return filekind.upper()

def locate_file(refname, mode=None):
    """Given a valid reffilename in CDBS or CRDS format,  return a cache path for the file.
    The aspect of this which is complicated is determining instrument and an instrument
    specific sub-directory for it based on the filename alone,  not the file contents.
    """
    _path,  _observatory, instrument, _filekind, _serial, _ext = get_reference_properties(refname)
    rootdir = locate_dir(instrument, mode)
    return  os.path.join(rootdir, os.path.basename(refname))

def locate_dir(instrument, mode=None):
    """Locate the instrument specific directory for a reference file."""
    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="tobs")
    else:
        config.check_crds_ref_subdir_mode(mode)
    crds_refpath = config.get_crds_refpath("tobs")
    prefix = get_env_prefix(instrument)
    if mode == "legacy":   # Locate cached files at the appropriate CDBS-style  iref$ locations
        try:
            rootdir = os.environ[prefix]
        except KeyError:
            try:
                rootdir = os.environ[prefix[:-1]]
            except KeyError as exc:
                raise KeyError(
                    "Reference location not defined for", repr(instrument),
                    ".  Did you configure", repr(prefix) + "?") from exc
    elif mode == "instrument":   # use simple names inside CRDS cache.
        rootdir = os.path.join(crds_refpath, instrument)
        refdir = os.path.join(crds_refpath, prefix[:-1])
        if not os.path.exists(refdir):
            if config.writable_cache_or_verbose("Skipping making instrument directory link for", repr(instrument)):
                log.verbose("Creating legacy cache link", repr(refdir), "-->", repr(rootdir))
                utils.ensure_dir_exists(rootdir + "/locate_dir.fits")
                os.symlink(rootdir, refdir)
    elif mode == "flat":    # use original flat cache structure,  all instruments in same directory.
        rootdir = crds_refpath
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return rootdir

# ============================================================================

def fits_to_parkeys(header):
    """Map a FITS header onto rmap parkeys appropriate for this observatory."""
    return dict(header)

# ============================================================================

def hijack_warnings(func, *args, **keys):
    """Re-map dependency warnings to CRDS warnings so they're counted and logged
    to web output.
    """
    with warnings.catch_warnings():
        # save and replace warnings.showwarning
        old_showwarning, warnings.showwarning = \
            warnings.showwarning, abstract.hijacked_showwarning

        # Always handle astropy warnings
        from astropy.utils.exceptions import AstropyUserWarning
        warnings.simplefilter("always", AstropyUserWarning)

        try:
            result = func(*args, **keys)
        finally:
            warnings.showwarning = old_showwarning

    return result

# =======================================================================

def test():
    """Run the module doctest_config."""
    import doctest
    from . import locate
    return doctest.testmod(locate)



#========================================================================
#
# module: siname
#
# Routines for determining the science instrument name variations
# and abbreviations. Includes:
#
# WhichInstrument()
# add_IRAF_prefix()
# get_ref_data_prefix()
# get_cdbs_prefix()
#
# History
# -------
# 11/05/02 xxxxx MSwam    first version
# 10/23/03 49468 MSwam    activate COS,WFC3 in CDBS
# 10/22/12  Todd Miller   hacked into CRDS
#
#========================================================================
import re

# exceptions
class UnknownInstrument(Exception):
  pass

class UnknownIRAFPrefix(Exception):
  pass

class UnknownRefDataPrefix(Exception):
  pass

class UnsupportedCDBSInstrument(Exception):
  pass

class UnknownCDBSPrefix(Exception):
  pass

# global compiled regex for splitting filenames
filesplit = re.compile("[_\.]")

# post-SM2 instruments supported by CDBS
CDBS_supports = ("ACS","STIS","NICMOS","WFPC2","WFC3","COS")

def WhichInstrument(dataset):
   """
=======================================================================
Name: WhichInstrument

Description:
------------
Determines instrument name given a dataset name in IPPPSSOOT format by
looking at the first character of the dataset name.

Inputs:  dataset (I) - dataset name in IPPPSSOOT format
Returns: a capitalized instrument name (e.g. ACS)

Exceptions: UnknownInstrument is thrown if the instrument cannot be
            determined.
History:
--------
10/01/02 xxxxx MSwam     Initial version
=======================================================================
   """
   try:
     return ID_CHAR_TO_INSTRUMENT[dataset[0].lower()]
   except:
     raise UnknownInstrument

ID_CHAR_TO_INSTRUMENT = {
  "j" : "ACS",
  "o": "STIS",
  "n": "NICMOS",
  "u": "WFPC2",
  "i": "WFC3",
  "l": "COS",
  "m" : "SYNPHOT",
  }

INSTRUMENT_TO_ID_CHAR = {
  val: key for (key, val) in ID_CHAR_TO_INSTRUMENT.items()
}

def instrument_to_id_char(instrument):
  """Given an instrument name (e.g. acs) return the corresponding CDBS
  instrument id character.
  """
  return INSTRUMENT_TO_ID_CHAR[instrument.upper()]

def add_IRAF_prefix(instrument_name):
    """
=======================================================================
Name: add_IRAF_prefix

Description:
------------
Determines IRAF prefix for a reference file name,
depending on instrument and reffile format.

Inputs:
-------
instrument_name - name of the instrument (e.g. ACS)

Returns:
--------
an iraf directory prefix (e.g. jref$ )

Exceptions: UnknownIRAFPrefix is thrown if the instrument cannot be
            determined.
History:
--------
10/01/02 xxxxx MSwam     Initial version
06/11/03 46981 MSwam     Don't use "tab" directories - all are "ref"
04/12/04 48141 MSwam     Add MULTI, SYNPHOT
=======================================================================
   """
    if (instrument_name == "ACS"):
      iraf_prefix = "jref$"
    elif (instrument_name == "STIS"):
      iraf_prefix = "oref$"
    elif (instrument_name == "NICMOS"):
      iraf_prefix = "nref$"
    elif (instrument_name == "WFPC2"):
      iraf_prefix = "uref$"
    elif (instrument_name == "WFC3"):
      iraf_prefix = "iref$"
    elif (instrument_name == "COS"):
      iraf_prefix = "lref$"
    elif (instrument_name == "MULTI"):
      iraf_prefix = "mtab$"
    elif (instrument_name == "SYNPHOT"):
      iraf_prefix = "ttab$"
    else:
      raise UnknownIRAFPrefix
    #
    return iraf_prefix

def get_ref_data_prefix(instrument_name):
    """
=======================================================================
Name: get_ref_data_prefix

Description:
------------
Translate instrument name into archive database "*_ref_data" prefix.
This is the 3-character prefix that is part of every field name in the
"*_ref_data" table.  For example, in stis_ref_data all fields are
prefixed with "ssr".

Inputs:
-------
instrument_name - name of the instrument (e.g. ACS)

Returns:
--------
an archive database "*_ref_data" prefix (e.g. for stis = ssr)

Exceptions: UnknownRefDataPrefix is thrown if the instrument cannot be
            determined.
History:
--------
10/01/02 xxxxx MSwam     Initial version
01/31/05 51433 MSwam     Change message to warning, since MULTI processing
                         results in a message for SM4 SIs
11/01/07 56836 Sherbert  Add WF3 and COS now that tables exist
=======================================================================
   """
    if (instrument_name == "ACS"):
      return "acr"
    elif (instrument_name == "STIS"):
      return "ssr"
    elif (instrument_name == "NICMOS"):
      return "nsr"
    elif (instrument_name == "WFPC2"):
      return "w2r"
    elif (instrument_name == "COS"):
      return "csr"
    elif (instrument_name == "WFC3"):
      return "w3r"
    else:
      raise UnknownRefDataPrefix

def supported_by_CDBS(instrument_name):
  """
=======================================================================
Name: supported_by_CDBS

Description:
------------
Throws an exception if the specified instrument is not yet supported
by CDBS.  The global string "CDBS_supports" contains the list of
instruments supported by CDBS.

Inputs:
-------
instrument_name - name of the instrument (e.g. ACS)

Exceptions: UnsupportedCDBSInstrument is raised if the instrument is
            not yet supported.
History:
--------
10/01/02 xxxxx MSwam     Initial version
=======================================================================
  """
  if not instrument_name in CDBS_supports:
    raise UnsupportedCDBSInstrument

def get_cdbs_prefix(instrument_name):
    """
=======================================================================
Name: get_cdbs_prefix

Description:
------------
Translate instrument name into CDBS database file and row table prefix.
Each HST instrument has two tables in CDBS, one for file-level info
(*_file) and one for row-level info (*_row).  Each has a prefix
indicating the instrument name (e.g. acs_row).

Inputs:
-------
instrument_name - name of the instrument (e.g. ACS)

Returns:
--------
the CDBS *_file and *_row table prefix for the instrument (e.g. acs)

Exceptions: UnknownCDBSPrefix is thrown if the instrument cannot be
            determined.
History:
--------
10/01/02 xxxxx MSwam     Initial version
=======================================================================
   """
    if instrument_name in CDBS_supports:
      return instrument_name.lower()
    elif (instrument_name == "SYNPHOT" or instrument_name == "MULTI"):
      return "synphot"
    else:
      raise UnknownCDBSPrefix

def WhichCDBSInstrument(reffilename):
   """
=======================================================================
Name: WhichCDBSInstrument

Description:
------------
Determines instrument name given a reference file name in CDBS format by
looking at the last character of the name, before the first
underscore.

Inputs:  reffilename (I) - dataset name in CDBS format
Returns: a capitalized instrument name (e.g. ACS)

NOTE: The reffilename can be the full FITS filename, or just the
      part of the filename included in the catalog entry.

Exceptions: UnknownInstrument is thrown if the instrument cannot be
            determined.
History:
--------
10/01/02 xxxxx MSwam     Initial version
05/03/05 51433 MSwam     Tweak syn test to work for catalog entries too
05/14/08 59478 MSwam     Add th (thermal) synphot type as well
=======================================================================
   """
   # split the filename into pieces on field delimiters
   parts = re.split(filesplit,reffilename)

   # first test for a synphot or thermal file (check the last two pieces)
   pos = len(parts) - 2
   if pos > 0 and parts[pos].lower() in ["syn", "th"]:
      return "SYNPHOT"
   pos += 1
   if pos > 0 and parts[pos].lower() in ["syn", "th"]:
      return "SYNPHOT"
   try:
     return ID_CHAR_TO_INSTRUMENT[parts[0][-1:].lower()]
   except:
      raise UnknownInstrument



from crds.list import ListScript
from crds.tests import test_config

"""
usage: crds list ...

usage: /Users/jmiller/work/workspace_crds/CRDS/crds/list.py
       [-h] [--references] [--mappings] [--cached-references]
       [--cached-mappings] [--cached-pickles] [--full-path]
       [--dataset-ids-for-instruments INSTRUMENTS [INSTRUMENTS ...]]
       [--dataset-headers IDS [IDS ...]] [--id-expansions-only]
       [--first-id-expansion-only] [--minimize-headers] [--json-headers]
       [--config] [--status] [--cat [FILES [FILES ...]]]
       [--keywords KEYWORDS [KEYWORDS ...]] [--add-filenames]
       [--operational-context] [--remote-context PIPELINE]
       [--resolve-contexts] [--required-parkeys]
       [--contexts [CONTEXT [CONTEXT ...]] | --range MIN:MAX | --all |
       --last-n-contexts N | --up-to-context CONTEXT | --after-context
       CONTEXT] [-v] [--verbosity VERBOSITY] [--dump-cmdline] [-R] [-I] [-V]
       [-J] [-H] [--stats] [--profile PROFILE] [--log-time] [--pdb]
       [--debug-traps]

crds.list prints out a variety of information about CRDS configuration, the
cache, reference and mapping files, default context names, and dataset headers
and ids used for CRDS reprocessing recommendations.


optional arguments:
  -h, --help            show this help message and exit
  --references          print names of reference files referred to by contexts
  --mappings            print names of mapping files referred to by contexts
  --cached-references   print the paths of all references in the local cache. very primitive.
  --cached-mappings     print the paths of all mappings in the local cache. very primitive.
  --cached-pickles      print the paths of all pickles in the local cache. very primitive.
  --full-path           print the full paths of listed files.
  --dataset-ids-for-instruments INSTRUMENTS [INSTRUMENTS ...]
                        print the dataset ids known to CRDS associated for the specified instruments.
  --dataset-headers IDS [IDS ...]
                        print matching parameters for the specified dataset ids.
  --id-expansions-only  print out only the <product>:<exposure> expansion associated with the specified --dataset-headers ids.
  --first-id-expansion-only
                        print out only the first exposure ID (header or expanded) associated with a particular product ID.
  --minimize-headers    print out only header parameters required by a particular CRDS context.
  --json-headers        print out header parameters in JSON format suited for crds.bestrefs and grepping.
  --config              print CRDS configuration information.
  --status              print brief, basic, CRDS configuration information.
  --cat [FILES [FILES ...]]
                        print the text of the specified mapping files.
  --keywords KEYWORDS [KEYWORDS ...]
                        limited list of keywords to be catted from reference headers.
  --add-filenames       prefix each line of a cat'ed file with the filename.
  --operational-context
                        print the name of the operational context on the central CRDS server.
  --remote-context PIPELINE
                        print the name of the context reported as in use by the specified pipeline.
  --resolve-contexts    print the literal names of the contexts defined by the command line context specifiers.
  --required-parkeys    print the names of the parkeys required to compute bestrefs for the specified mappings.
  --contexts [CONTEXT [CONTEXT ...]]
                        Specify a list of CRDS mappings to operate on: .pmap, .imap, or .rmap or date-based specification
  --range MIN:MAX       Operate for pipeline context ids (.pmaps) between <MIN> and <MAX>.
  --all                 Operate with respect to all known CRDS contexts.
  --last-n-contexts N   Operate with respect to the last N contexts.
  --up-to-context CONTEXT
                        Operate on all contexts up to and including the specified context.
  --after-context CONTEXT
                        Operate on all contexts after and including the specified context.
  -v, --verbose         Set log verbosity to True,  nominal debug level.
  --verbosity VERBOSITY
                        Set log verbosity to a specific level: 0..100.
  --dump-cmdline        Dump the command line parameters used to start the script to the log.
  -R, --readonly-cache  Don't modify the CRDS cache.  Not compatible with options which implicitly modify the cache.
  -I, --ignore-cache    Download required files even if they're already in the cache.
  -V, --version         Print the software version and exit.
  -J, --jwst            Force observatory to JWST for determining header conventions.
  -H, --hst             Force observatory to HST for determining header conventions.
  --stats               Track and print timing statistics.
  --profile PROFILE     Output profile stats to the specified file.
  --log-time            Add date/time to log messages.
  --pdb                 Run under pdb.
  --debug-traps         Bypass exception error message traps and re-raise exception.

    General categories of information driven by switches include:

    0. Overall CRDS configuration
    1. CRDS server file lists
    2. CRDS cache file lists and paths
    3. Cached file contents or headers
    4. CRDS reprocessing dataset ids and parameters
    5. Listing global default and installed pipeline contexts
    6. Resolving context specifiers into literal context names

    --------------------------------------------------------------------------
    0. Configuration information governing the behavior of CRDS for simple
    configurations can be dumped:

    % crds list --status
    CRDS Version = '7.0.7, bump-version, 7432326'
    CRDS_MODE = 'auto'
    CRDS_PATH = '/Users/jmiller/crds_cache_ops'
    CRDS_SERVER_URL = 'https://jwst-crds.stsci.edu'
    Effective Context = 'jwst_0204.pmap'
    Last Synced = '2016-09-20 08:00:09.115330'
    Python Executable = '/Users/jmiller/anaconda/bin/python'
    Python Version = '3.5.2.final.0'
    Readonly Cache = False

    More comprehensive configuration information is also available for advanced
    configurations:

    % crds list --config
    ... lots of info ....

    --------------------------------------------------------------------------
   1. Files known by the CRDS server to belong to specified contexts can be listed
    even if the files are not installed in a local CRDS Cache.

    The --mappings command recursively evaluates and includes all the sub-mappings,
    i.e. imaps and pmaps, of the specified contexts.

    Contexts to list can be specified in a variety of ways:

    -- To list the references contained by several contexts

    % crds list  --references --contexts hst_0001.pmap hst_0002.pmap ...
    vb41935ij_bia.fits
    vb41935kj_bia.fits
    ...

    -- To list the references in a numerical range of contexts

    % crds list --references --range 1:2 --references
    vb41935lj_bia.fits
    vb41935oj_bia.fits
    ...

    -- To list all mappings, even those not referenced by an imap or pmap

    % crds list --mappings --all
    hst.pmap
    hst_0001.pmap
    hst_0002.pmap
    hst_acs.imap
    hst_acs_0001.imap
    hst_acs_0002.imap
    hst_acs_atodtab.rmap
    ...

    --references, --mappings, or both can be listed.

    --------------------------------------------------------------------------
    2. Locally cached files (files already synced to your computer) can be listed:

    % crds list --cached-mappings --full-path
    ...

    % crds list --cached-references --full-path
    ...

    In both cases adding --full-path prints the path of the file within the CRDS cache.

    These are merely simple directory listings which ignore the context specifiers
    and can be grep'ed for finer grained answers.

    --------------------------------------------------------------------------
    3. The contents of cached mappings or references (header only) can be printed to stdout like this:

    % crds list --contexts jwst-fgs-linearity-edit jwst-nirspec-linearity-edit --cat --add-filename | grep parkey
    CRDS - INFO - Symbolic context 'jwst-fgs-linearity-edit' resolves to 'jwst_fgs_linearity_0008.rmap'
    CRDS - INFO - Symbolic context 'jwst-nirspec-linearity-edit' resolves to 'jwst_nirspec_linearity_0009.rmap'
    /cache/path/mappings/jwst/jwst_fgs_linearity_0008.rmap:     'parkey' : (('META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),
    /cache/path/mappings/jwst/jwst_nirspec_linearity_0009.rmap:     'parkey' : (('META.INSTRUMENT.DETECTOR', 'META.SUBARRAY.NAME'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),

    this prints the contents of the specified rmaps.

    The -edit specifier above refers to mappings contained by the default starting point (.pmap) of future
    server submissions.  It tracks on-going submission work that precedes the adoption of a new context
    as the default in use by the pipeline.

    crds.list --cat can be applied to references and prints out the reference metadata that CRDS views
    abstractly as the file header.

    References need to be catted explicitly by name,  but the list can come from the --references command
    explained above:

    % crds list --cat jwst_nirspec_dark_0036.fits
    CRDS - INFO - Symbolic context 'jwst-operational' resolves to 'jwst_0167.pmap'
    ##########################################################################################
    File:  '.../references/jwst/jwst_nirspec_dark_0036.fits'
    ##########################################################################################
    {'A1_COL_C': '8.9600000e+002',
    'A1_CONF1': '2.1846000e+004',
    ...
    }

   --------------------------------------------------------------------------
   4. Information about the dataset IDs and their associated parameters used
   for CRDS reprocessing can be printed:

    % crds list --dataset-headers jcl403010 --first-id --minimize-header
    CRDS - INFO - Symbolic context 'hst-operational' resolves to 'hst_0462.pmap'
    CRDS - INFO - Dataset pars for 'JCL403010:JCL403ECQ' with respect to 'hst_0462.pmap':
    {'APERTURE': 'WFC1',
     'ATODCORR': 'OMIT',
     'BIASCORR': 'COMPLETE',
     'CCDAMP': 'ABCD',
     'CCDCHIP': '-999.0',
     'CCDGAIN': '2.0',
     'CRCORR': 'OMIT',
     'DARKCORR': 'COMPLETE',
     'DATE-OBS': '2016-02-20',
     'DETECTOR': 'WFC',
     'DQICORR': 'COMPLETE',
     'DRIZCORR': 'COMPLETE',
     'FILTER1': 'CLEAR1L',
     'FILTER2': 'F814W',
     'FLASHCUR': 'LOW',
     'FLATCORR': 'COMPLETE',
     'FLSHCORR': 'OMIT',
     'FW1OFFST': '0.0',
     'FW2OFFST': '0.0',
     'FWSOFFST': '0.0',
     'GLINCORR': 'UNDEFINED',
     'INSTRUME': 'ACS',
     'LTV1': '0.0',
     'LTV2': '0.0',
     'NAXIS1': '4144.0',
     'NAXIS2': '4136.0',
     'OBSTYPE': 'IMAGING',
     'PCTECORR': 'UNDEFINED',
     'PHOTCORR': 'COMPLETE',
     'RPTCORR': 'UNDEFINED',
     'SHADCORR': 'OMIT',
     'SHUTRPOS': 'A',
     'TIME-OBS': '17:32:29.666665',
     'XCORNER': '0.0',
     'YCORNER': '0.0',
     'dataset_id': 'JCL403010:JCL403ECQ'}

    Sometimes it's desirable to know the individual exposures CRDS associates with a product id:

    % crds list --dataset-headers jcl403010 --id-expansions-only
    CRDS - INFO - Symbolic context 'hst-operational' resolves to 'hst_0462.pmap'
    JCL403010:JCL403ECQ
    JCL403010:JCL403EEQ
    JCL403010:JCL403EGQ
    JCL403010:JCL403EIQ
    JCL403010:JCL403EKQ
    JCL403010:JCL403EMQ
    JCL403010:JCL403EOQ
    JCL403010:JCL403EQQ
    JCL403010:JCL403ESQ
    JCL403010:JCL403EUQ

    Headers available can possibly vary by CRDS context and will be dumped for
    every specified or implicit context.  Generally the default context is
    sufficient.  Often all exposures of an association have identical
    parameters but CRDS is designed so that this does not have to be the case.

    These dataset header services require setting CRDS_SERVER_URL to a valid CRDS server to
    provide a source for the headers.

   --------------------------------------------------------------------------
    5. Information about the default context can be printed.  There are two variations and a subtle distinction:

    % python m crds.list --operational-context
    jwst_0204.pmap

    lists the context which has been *commanded* as default on the CRDS server.

    While:

    % crds list --remote-context jwst-ops-pipeline
    jwst_0101.pmap

    lists the context which is *in actual use* in the associated archive pipeline as reported by
    a cache sync echo.

    During the interval between commanding a new default on the CRDS server and syncing the pipeline
    CRDS cache,  the commanded and actual pipeline contexts can differ.

   --------------------------------------------------------------------------
    6. Resolving context specifiers

    Some CRDS tools, including crds.list and crds.sync, support multiple
    mechanisms for specifying context.  The --resolve-contexts command
    interprets those specifiers into a non-recursive list of literal mapping
    names and prints them out.  --resolve-contexts differs from --mappings
    because it does not implicitly include all sub-mappings of the specified
    contexts.

    % crds list --resolve-contexts --all
    jwst.pmap
    jwst_0000.pmap
    jwst_0001.pmap
    jwst_0002.pmap
    jwst_0003.pmap
    ...

    % crds list --resolve-contexts --last 5
    jwst_0205.pmap
    jwst_0206.pmap
    jwst_0207.pmap
    jwst_0208.pmap
    jwst_0209.pmap

    % crds list --resolve-contexts  --contexts jwst-miri-dark-operational
    jwst_miri_dark_0012.rmap

    % crds list --resolve-contexts --contexts jwst-niriss-superbias-2016-01-01T00:00:00
    jwst_niriss_superbias_0005.rmap


----------
TEST CASES
----------
"""

from crds.core import log
from crds import tests
from crds.tests import test_config

from crds.diff import DiffScript

def dt_list_mappings():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --mappings --contexts hst.pmap")() # doctest: +ELLIPSIS
    hst.pmap
    hst_acs.imap
    hst_acs_atodtab.rmap
    hst_acs_biasfile.rmap
    hst_acs_bpixtab.rmap
    hst_acs_ccdtab.rmap
    hst_acs_cfltfile.rmap
    hst_acs_crrejtab.rmap
    hst_acs_d2imfile.rmap
    hst_acs_darkfile.rmap
    hst_acs_dgeofile.rmap
    hst_acs_drkcfile.rmap
    hst_acs_flshfile.rmap
    ...
    hst_wfpc2_shadfile.rmap
    hst_wfpc2_wf4tfile.rmap
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_cached_mappings():
    """
    >>> old_state = test_config.setup()
    >>> import doctest
    >>> doctest.ELLIPSIS_MARKER = '-ignore-'
    >>> ListScript("crds.list --cached-mappings --full-path")() # doctest: +ELLIPSIS
    -ignore-/mappings/hst/hst.pmap
    -ignore-/mappings/hst/hst_0001.pmap
    -ignore-/mappings/hst/hst_0002.pmap
    -ignore-
    0
    >>> doctest.ELLIPSIS_MARKER = '...'
    >>> test_config.cleanup(old_state)
    """

def dt_list_references():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --references --contexts hst.pmap")() # doctest: +ELLIPSIS
    d6n1328ou.r6h
    d6n13293u.r6h
    d6n13298u.r6h
    d6n1329fu.r6h
    d6n1329qu.r6h
    ...
    x781756sj_dkc.fits
    x781933ij_drk.fits
    x781933jj_dkc.fits
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_cached_references():
    """
    >>> import doctest
    >>> doctest.ELLIPSIS_MARKER = '-ignore-'
    >>> old_state = test_config.setup(cache=test_config.CRDS_TESTING_CACHE, observatory="hst")
    >>> ListScript("crds.list --cached-references --full-path")() # doctest: +ELLIPSIS
    -ignore-/crds-cache-test/references/hst/v8q14451j_idc.fits
    -ignore-/crds-cache-test/references/hst/z1q2219el_2zx.fits
    -ignore-/crds-cache-test/references/hst/z1r1943el_1dx.fits
    -ignore-/crds-cache-test/references/hst/z2d19237l_1dx.fits
    -ignore-/crds-cache-test/references/hst/z2d1925ql_2zx.fits
    0
    >>> test_config.cleanup(old_state)
    >>> doctest.ELLIPSIS_MARKER = '...'
    """

def dt_list_dataset_ids():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --dataset-ids wfpc2")()   # doctest: +ELLIPSIS
    U20L0U01T:U20L0U01T
    U20L0U02T:U20L0U02T
    U20L0U03T:U20L0U03T
    U21U0201T:U21U0201T
    ...
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_dataset_headers():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --dataset-headers U20L0U01T:U20L0U01T --minimize-header --contexts hst.pmap")()   # doctest: +ELLIPSIS
    Dataset pars for 'U20L0U01T:U20L0U01T' with respect to 'hst.pmap':
     {'ATODGAIN': '15.0',
     'CRDS_CTX': 'hst.pmap',
     'DATE-OBS': '1993-12-07',
     'FILTER1': '0.0',
     'FILTER2': '0.0',
     'FILTNAM1': 'UNDEFINED',
     'FILTNAM2': 'UNDEFINED',
     'IMAGETYP': 'BIAS',
     'INSTRUME': 'WFPC2',
     'LRFWAVE': '0.0',
     'MODE': 'FULL',
     'SERIALS': 'OFF',
     'SHUTTER': 'A',
     'TIME-OBS': '10:07:16.929999',
     'dataset_id': 'U20L0U01T:U20L0U01T'}
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_dataset_headers_json():
    """
    >>> old_state = test_config.setup()
    >>> import doctest
    >>> doctest.ELLIPSIS_MARKER = '-ignore-'
    >>> ListScript("crds.list --dataset-headers U20L0U01T:U20L0U01T --contexts hst.pmap --json")()   # doctest: +ELLIPSIS
    -ignore-
    0
    >>> test_config.cleanup(old_state)
    >>> doctest.ELLIPSIS_MARKER = '...'
    """

def dt_list_dataset_headers_bogus():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --dataset-headers BAR:BAR --contexts hst.pmap")()   # doctest: +ELLIPSIS
    CRDS - ERROR -  Failed fetching dataset parameters with repect to 'hst.pmap' for ['BAR:BAR'] : CRDS jsonrpc failure 'get_dataset_headers_by_id' OtherError: Can't determine instrument for dataset ...'BAR'
    1
    >>> test_config.cleanup(old_state)
    """

def dt_list_dataset_headers_id_expansions_only():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --dataset-headers I9ZF01010 --id-expansions-only --contexts hst.pmap")()   # doctest: +ELLIPSIS
    I9ZF01010:I9ZF01DZQ
    I9ZF01010:I9ZF01E0Q
    I9ZF01010:I9ZF01E1Q
    I9ZF01010:I9ZF01E3Q
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_required_parkeys_pmap():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --required-parkeys --contexts hst.pmap")() # doctest: +ELLIPSIS
    --------------------- Parkeys required for 'hst.pmap' ---------------------
    acs = ['INSTRUME', 'APERTURE', 'ATODCORR', 'BIASCORR', 'CCDAMP', 'CCDCHIP', 'CCDGAIN', 'CRCORR', 'DARKCORR', 'DATE-OBS', 'DETECTOR', 'DQICORR', 'DRIZCORR', 'FILTER1', 'FILTER2', 'FLASHCUR', 'FLATCORR', 'FLSHCORR', 'FW1OFFST', 'FW2OFFST', 'FWSOFFST', 'GLINCORR', 'LTV1', 'LTV2', 'NUMCOLS', 'NUMROWS', 'OBSTYPE', 'PCTECORR', 'PHOTCORR', 'REFTYPE', 'SHADCORR', 'SHUTRPOS', 'TIME-OBS', 'XCORNER', 'YCORNER']
    cos = ['INSTRUME', 'BADTCORR', 'BRSTCORR', 'DATE-OBS', 'DEADCORR', 'DETECTOR', 'EXPTYPE', 'FLATCORR', 'FLUXCORR', 'LIFE_ADJ', 'OBSMODE', 'OBSTYPE', 'OPT_ELEM', 'REFTYPE', 'TDSCORR', 'TIME-OBS', 'WALKCORR']
    nicmos = ['INSTRUME', 'CAMERA', 'DATE-OBS', 'FILTER', 'NREAD', 'OBSMODE', 'READOUT', 'REFTYPE', 'SAMP_SEQ', 'TIME-OBS']
    stis = ['INSTRUME', 'APERTURE', 'BINAXIS1', 'BINAXIS2', 'CCDAMP', 'CCDGAIN', 'CCDOFFST', 'CENWAVE', 'DATE-OBS', 'DETECTOR', 'OBSTYPE', 'OPT_ELEM', 'REFTYPE', 'TIME-OBS']
    wfc3 = ['INSTRUME', 'APERTURE', 'ATODCORR', 'BIASCORR', 'BINAXIS1', 'BINAXIS2', 'CCDAMP', 'CCDGAIN', 'CHINJECT', 'DARKCORR', 'DATE-OBS', 'DETECTOR', 'DQICORR', 'DRIZCORR', 'FILTER', 'FLASHCUR', 'FLATCORR', 'FLSHCORR', 'PHOTCORR', 'REFTYPE', 'SAMP_SEQ', 'SHUTRPOS', 'SUBARRAY', 'SUBTYPE', 'TIME-OBS']
    wfpc2 = ['INSTRUME', 'ATODGAIN', 'DATE-OBS', 'FILTER1', 'FILTER2', 'FILTNAM1', 'FILTNAM2', 'IMAGETYP', 'LRFWAVE', 'MODE', 'REFTYPE', 'SERIALS', 'SHUTTER', 'TIME-OBS']
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_required_parkeys_imap():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --required-parkeys --contexts hst_acs.imap")() # doctest: +ELLIPSIS
    atodtab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'ATODCORR']
    biasfile: ['DETECTOR', 'CCDAMP', 'CCDGAIN', 'APERTURE', 'NUMCOLS', 'NUMROWS', 'LTV1', 'LTV2', 'XCORNER', 'YCORNER', 'CCDCHIP', 'DATE-OBS', 'TIME-OBS', 'BIASCORR', 'XCORNER', 'YCORNER', 'CCDCHIP']
    bpixtab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'DQICORR']
    ccdtab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS']
    cfltfile: ['DETECTOR', 'FILTER1', 'FILTER2', 'OBSTYPE', 'DATE-OBS', 'TIME-OBS']
    crrejtab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'CRCORR']
    d2imfile: ['DATE-OBS', 'TIME-OBS', 'DRIZCORR']
    darkfile: ['DETECTOR', 'CCDAMP', 'CCDGAIN', 'DATE-OBS', 'TIME-OBS', 'DARKCORR']
    dgeofile: ['DETECTOR', 'FILTER1', 'FILTER2', 'DATE-OBS', 'TIME-OBS']
    drkcfile: ['DETECTOR', 'CCDAMP', 'CCDGAIN', 'DATE-OBS', 'TIME-OBS', 'PCTECORR']
    flshfile: ['DETECTOR', 'CCDAMP', 'CCDGAIN', 'FLASHCUR', 'SHUTRPOS', 'DATE-OBS', 'TIME-OBS', 'FLSHCORR']
    idctab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS']
    imphttab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'PHOTCORR']
    mdriztab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'DRIZCORR']
    mlintab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'GLINCORR']
    npolfile: ['DETECTOR', 'FILTER1', 'FILTER2', 'DATE-OBS', 'TIME-OBS', 'DRIZCORR']
    oscntab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS']
    pctetab: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'PCTECORR']
    pfltfile: ['DETECTOR', 'CCDAMP', 'FILTER1', 'FILTER2', 'OBSTYPE', 'FW1OFFST', 'FW2OFFST', 'FWSOFFST', 'DATE-OBS', 'TIME-OBS', 'FLATCORR']
    shadfile: ['DETECTOR', 'DATE-OBS', 'TIME-OBS', 'SHADCORR']
    spottab: ['DETECTOR', 'OBSTYPE', 'DATE-OBS', 'TIME-OBS']
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_required_parkeys_rmap():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --required-parkeys --contexts hst_acs_darkfile.rmap")() # doctest: +ELLIPSIS
    hst_acs_darkfile.rmap: ['DETECTOR', 'CCDAMP', 'CCDGAIN', 'DATE-OBS', 'TIME-OBS', 'DARKCORR']
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_resolve_contexts_range():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --resolve-contexts --range 0:5")() # doctest: +ELLIPSIS
    hst_0001.pmap
    hst_0002.pmap
    hst_0003.pmap
    hst_0004.pmap
    hst_0005.pmap
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_resolve_contexts_date():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --resolve-contexts --contexts hst-2014-11-11T00:00:00")() # doctest: +ELLIPSIS
    hst_0297.pmap
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_remote_context():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --remote-context hst-ops-pipeline")() # doctest: +ELLIPSIS
    hst_....pmap
    >>> test_config.cleanup(old_state)
    """

def dt_list_operational_context():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --operational-context")() # doctest: +ELLIPSIS
    hst_....pmap
    >>> test_config.cleanup(old_state)
    """

def dt_list_references_by_context():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --references --contexts hst-cos-deadtab-2014-11-11T00:00:00")() # doctest: +ELLIPSIS
    s7g1700gl_dead.fits
    s7g1700ql_dead.fits
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_references_by_context():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --references --contexts hst-cos-deadtab-2014-11-11T00:00:00")() # doctest: +ELLIPSIS
    s7g1700gl_dead.fits
    s7g1700ql_dead.fits
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_cat_mappings():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --cat --mappings --contexts hst-cos-deadtab-2014-11-11T00:00:00")() # doctest: +ELLIPSIS
    ################################################################################
    File: .../mappings/hst/hst_cos_deadtab_0250.rmap
    --------------------------------------------------------------------------------
    header = {
        'derived_from' : 'generated from CDBS database 2014-05-09 23:24:57.840119',
        'filekind' : 'DEADTAB',
        'instrument' : 'COS',
        'mapping' : 'REFERENCE',
        'name' : 'hst_cos_deadtab_0250.rmap',
        'observatory' : 'HST',
        'parkey' : (('DETECTOR',), ('DATE-OBS', 'TIME-OBS')),
        'reffile_format' : 'TABLE',
        'reffile_required' : 'NONE',
        'reffile_switch' : 'DEADCORR',
        'rmap_relevance' : '(DEADCORR != "OMIT")',
        'sha1sum' : 'bde314f1848b67891d6309b30eaa5c95611f86e2',
    }
    <BLANKLINE>
    selector = Match({
        ('FUV',) : UseAfter({
            '1996-10-01 00:00:00' : 's7g1700gl_dead.fits',
        }),
        ('NUV',) : UseAfter({
            '1996-10-01 00:00:00' : 's7g1700ql_dead.fits',
        }),
    })
    --------------------------------------------------------------------------------
    Catalog Info:
    ................................................................................
    {'activation_date': '2014-05-15 15:31:54',
     'aperture': 'none',
     'blacklisted': 'false',
     'change_level': 'severe',
     'comment': 'none',
     'creator_name': 'todd miller',
     'delivery_date': '2014-05-11 08:14:02',
     'derived_from': 'generated from cdbs database 2014-05-09 23:24:57.840119',
     'description': 'rebaselined hst rmaps as 250-series for opus 2014.2',
     'filekind': 'deadtab',
     'history': 'none',
     'instrument': 'cos',
     'name': 'hst_cos_deadtab_0250.rmap',
     'observatory': 'hst',
     'pedigree': '',
     'reference_file_type': '',
     'rejected': 'false',
     'replaced_by_filename': '',
     'sha1sum': '41cbdf620d41586fbc3de3e26d14d14eb42cc244',
     'size': '711',
     'state': 'operational',
     'type': 'mapping',
     'uploaded_as': 'hst_cos_deadtab_0250.rmap',
     'useafter_date': '2050-01-01 00:00:00'}

    >>> test_config.cleanup(old_state)
    """


def dt_list_status():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --status")() # doctest: +ELLIPSIS
    CRDS Summary = '...'
    CRDS Version = '...'
    CRDS_MODE = 'auto'
    CRDS_PATH = '...'
    CRDS_SERVER_URL = 'https://....stsci.edu'
    Cache Locking = 'enabled, multiprocessing'
    Effective Context = 'hst_....pmap'
    Last Synced = '...'
    Python Executable = '...'
    Python Version = '...'
    Readonly Cache = False
    0
    >>> test_config.cleanup(old_state)
    """

def dt_list_config():
    """
    >>> old_state = test_config.setup()
    >>> ListScript("crds.list --config")() # doctest: +ELLIPSIS
    CRDS Environment
    ...
    CRDS Client Config
    ...
    CRDS Actual Paths
    ...
    CRDS Server Info
    ...
    Calibration Environment
    ...
    Python Environment
    ...
    0
    >>> test_config.cleanup(old_state)
    """

def main():
    """Run module tests,  for now just doctests only.

    test_config.setup() and cleanup() are done inline above because bracketing
    the tests here does not get picked up by nose test discovery.  Combining
    tests into one giant docstring works but is hard to analyze and debug when
    things go wrong.
    """
    from crds.tests import test_list, tstmod
    return tstmod(test_list)

if __name__ == "__main__":
    print(main())
